import pytest
import copy
from railtracks.llm.models._litellm_wrapper import _parameters_to_json_schema
from railtracks.llm.tools.schema_parser import parse_json_schema_to_parameter
import json
from .conftest import load_json_schemas, deep_equal

schemas = load_json_schemas()

@pytest.mark.parametrize("input_schema", [pytest.param(s, id=f"Case {i}") for i, s in enumerate(schemas)])
def test_jsonschema_roundtrip(input_schema):
    # Deepcopy to avoid mutation during test
    schema = copy.deepcopy(input_schema)

    # Step 1: Use parser to get parameter objects (algorithm below)
    # Extract params from top-level props, required, etc.
    parameters = []
    properties = schema.get("properties", {})
    required = set(schema.get("required", []))
    for name, prop_schema in properties.items():
        param = parse_json_schema_to_parameter(
            name=name,
            prop_schema=prop_schema,
            required=name in required
        )
        parameters.append(param)

    # make a param set 
    params = list(parameters) 

    # Step 2: Convert back to schema
    regenerated = _parameters_to_json_schema(params)
    # regenerated.pop("additionalProperties", None)   # this is needed by litellm

    schema_ = copy.deepcopy(schema)
    regenerated_ = copy.deepcopy(regenerated)
    schema_.pop("$schema", None)
    regenerated_.pop("$schema", None)

    ######## INCASE YOU WANT TO DEBUG ########
    # print(json.dumps(schema_, indent=2))
    # print(json.dumps(regenerated_, indent=2))
    ##########################################

    # Step 3: Deep equality check
    assert deep_equal(schema_, regenerated_), "Schemas are not deeply equal"



import time

import pytest
from railtracks.state.request import (
    RequestTemplate, RequestForest, RequestDoesNotExistError,
    RequestAlreadyExistsError, Cancelled, Failure,
)
from railtracks.utils.profiling import Stamp



# ================ START RequestTemplate unit tests ======================
def test_repr_and_properties(req_template_factory):
    r = req_template_factory(output=None)
    assert "RequestTemplate" in repr(r)
    assert r.closed is False
    assert r.status == "Open"
    assert r.is_insertion is (r.source_id is not None or False)
    # Set output to close
    completed = req_template_factory(output="OK!", parent=r)
    assert completed.closed is True
    assert completed.status == "Completed"
    assert completed.parent == r

def test_to_edge_and_hierarchy(req_template_factory):
    parent = req_template_factory(step=2, output=123)
    child = req_template_factory(step=3, parent=parent, output=None)
    edge = child.to_edge()
    assert edge.source == child.source_id
    assert edge.target == child.sink_id
    assert edge.details["status"] == child.status
    assert edge.details["output"] == child.output
    assert edge.parent is not None
    assert edge.parent.identifier == parent.identifier

def test_get_all_parents_and_terminal_parent(req_template_factory):
    leaf = req_template_factory(step=3)
    mid = req_template_factory(step=2, parent=leaf)
    root = req_template_factory(step=1, parent=mid)
    parents = root.get_all_parents()
    assert parents == [root, mid, leaf]
    t = root.get_terminal_parent
    assert t == leaf

def test_duration_detail(req_template_factory, req_stamp):
    leaf = req_template_factory(step=1)
    mid = req_template_factory(step=5, parent=leaf)
    assert mid.duration_detail == req_stamp(5).time - req_stamp(1).time

def test_generate_id_unique():
    ids = {RequestTemplate.generate_id() for _ in range(100)}
    assert len(ids) == 100

def test_children_complete_and_downstream_upstream(req_template_factory):
    r1 = req_template_factory(source_id=None, sink_id="A", output=1, step=1)
    r2 = req_template_factory(source_id="A", sink_id="B", output=None, step=2)
    r3 = req_template_factory(source_id="A", sink_id="C", output="done", step=3)
    requests = [r1, r2, r3]
    # children_complete
    assert not RequestTemplate.children_complete(requests, "A")
    # close all, should be true:
    r2c = req_template_factory(source_id="A", sink_id="B", output="x", step=4)
    assert RequestTemplate.children_complete([r1, r2c, r3], "A")
    # downstream: r2, r3 both with source_id="A"
    ds = RequestTemplate.downstream(requests, "A")
    assert {r.sink_id for r in ds} == {"B", "C"}
    # upstream: r1 with sink_id="A"
    us = RequestTemplate.upstream(requests, "A")
    assert us[0].sink_id == "A"

def test_all_downstream_collects_recursively(req_template_factory):
    # A -> B, A -> C, B -> D, C -> E
    A = req_template_factory(source_id=None, sink_id="A", step=1)
    B = req_template_factory(source_id="A", sink_id="B", step=2)
    C = req_template_factory(source_id="A", sink_id="C", step=2)
    D = req_template_factory(source_id="B", sink_id="D", step=3)
    E = req_template_factory(source_id="C", sink_id="E", step=4)
    requests = [A, B, C, D, E]
    results = RequestTemplate.all_downstream(requests, "A")
    assert {x.sink_id for x in results} == {"B", "C", "D", "E"}
# ================ END RequestTemplate unit tests ========================

# =============== START open_tails structure tests ===============
def test_open_tails_starting_from_subtree(req_template_factory):
    # Build a chain F->G->H (all open), only H should be the open tail
    F = req_template_factory(source_id=None, sink_id="F", output=None, step=1)
    G = req_template_factory(source_id="F", sink_id="G", output=None, step=2)
    H = req_template_factory(source_id="G", sink_id="H", output=None, step=3)
    all_reqs = [F, G, H]
    tails = RequestTemplate.open_tails(all_reqs, "F")
    assert len(tails) == 1
    assert tails[0].sink_id == "H"
    # If we close H, then G becomes tail
    Hc = req_template_factory(source_id="G", sink_id="H", output="y", step=3)
    all_reqs = [F, G, Hc]
    tails2 = RequestTemplate.open_tails(all_reqs, "F")
    assert len(tails2) == 1 and tails2[0].sink_id == "G"

def test_open_tails_no_open_leaves(req_template_factory):
    I = req_template_factory(source_id=None, sink_id="I", output="closed", step=1)
    assert RequestTemplate.open_tails([I], None) == []

def test_open_tails_deep_structure(req_template_factory):
    # Deep chain with one open in the middle
    J = req_template_factory(source_id=None, sink_id="J", output=None, step=1)
    K = req_template_factory(source_id="J", sink_id="K", output=None, step=2)
    L = req_template_factory(source_id="K", sink_id="L", output="closed", step=3)
    all_reqs = [J, K, L]
    tails = RequestTemplate.open_tails(all_reqs, None)
    # Should return only K (since it's open, child L is closed)
    assert [t.sink_id for t in tails] == ["K"]

def test_open_tails_multiple_starting_points(req_template_factory):
    # Multiple roots
    M = req_template_factory(source_id=None, sink_id="M", output=None, step=1)
    N = req_template_factory(source_id=None, sink_id="N", output="done", step=2)
    res = RequestTemplate.open_tails([M, N], None)
    assert len(res) == 1 and res[0].sink_id == "M"
# =============== END open_tails structure tests ===============

# ================ START RequestForest tests =============================
def test_forest_create_and_prevents_duplicate(req_forest, req_stamp):
    rid = RequestTemplate.generate_id()
    req_forest.create(rid, source_id=None, sink_id='A', input_args=(), input_kwargs={}, stamp=req_stamp(0))
    # Should raise on duplicate
    with pytest.raises(RequestAlreadyExistsError):
        req_forest.create(rid, source_id=None, sink_id='A', input_args=(), input_kwargs={}, stamp=req_stamp(1))

def test_forest_update_and_not_exists(req_forest, req_template_factory, req_stamp):
    # Create and then update
    r = req_template_factory()
    rid = r.identifier
    req_forest._update_heap(r)
    req_forest.update(rid, output="finished!", stamp=req_stamp(1))
    assert req_forest[rid].output == "finished!"
    # Not in heap
    with pytest.raises(RequestDoesNotExistError):
        req_forest.update("nope", output="?", stamp=req_stamp(1))

def test_forest_children_and_children_requests_complete(req_forest, req_template_factory):
    A = req_template_factory(identifier="A", source_id=None, sink_id="A", output="ins", step=1)
    B = req_template_factory(identifier="B", source_id="A", sink_id="B", output="ok", step=2)
    C = req_template_factory(identifier="C", source_id="A", sink_id="C", output="ok", step=3)
    req_forest._update_heap(A)
    req_forest._update_heap(B)
    req_forest._update_heap(C)
    # Should find both B and C as children of A
    children = req_forest.children("A")
    assert set(x.sink_id for x in children) == {"B", "C"}
    # Check children_requests_complete
    got = req_forest.children_requests_complete("A")
    assert got == A.identifier

def test_forest_get_request_from_child_id(req_forest, req_template_factory):
    A = req_template_factory(identifier="A", source_id=None, sink_id="A", output="ins", step=1)
    B = req_template_factory(identifier="B", source_id="A", sink_id="B", output=None, step=2)
    for req in (A, B):
        req_forest._update_heap(req)
    returned = req_forest.get_request_from_child_id("A")
    assert returned == A
    # Not present: should raise
    with pytest.raises(RequestDoesNotExistError):
        req_forest.get_request_from_child_id("X")

def test_forest_open_tails_and_insertion_request(req_forest, req_template_factory):
    A = req_template_factory(identifier="A", source_id=None, sink_id="A", output=None, step=1)
    B = req_template_factory(identifier="B", source_id="A", sink_id="B", output="out", step=2)
    req_forest._update_heap(A)
    req_forest._update_heap(B)
    # Open tails should return A
    tails = req_forest.open_tails()
    assert len(tails) == 1 and tails[0].sink_id == "A"
    # insertion_request returns all with source_id is None
    ins = req_forest.insertion_request
    assert ins and ins[0].sink_id == "A"
    assert req_forest.answer == None # answer property retrieves insertion output

def test_forest_answer_handles_no_insertions(req_forest):
    # Heap is empty
    assert req_forest.answer is None

def test_forest_to_edges_returns_all(req_forest, req_template_factory):
    A = req_template_factory(identifier="A", source_id=None, sink_id="A", output="foo", step=1)
    B = req_template_factory(identifier="B", source_id="A", sink_id="B", step=2, output="bar")
    req_forest._update_heap(A)
    req_forest._update_heap(B)
    edge_list = req_forest.to_edges()
    assert len(edge_list) == 2
    assert {e.source for e in edge_list} == {None, "A"}

def test_forest_generate_graph(req_template_factory):
    # Build three requests: None->A, A->B, A->C
    A = req_template_factory(identifier="A", source_id=None, sink_id="A", step=1)
    B = req_template_factory(identifier="B", source_id="A", sink_id="B", step=2)
    C = req_template_factory(identifier="C", source_id="A", sink_id="C", step=3)
    heap = {r.identifier: r for r in [A, B, C]}
    graph = RequestForest.generate_graph(heap)
    # Edges from None to A (insertion), and from A to B, A to C
    assert (("A", A.identifier) in graph[None])
    assert (("B", B.identifier) in graph["A"])
    assert (("C", C.identifier) in graph["A"])
    # Each node explicitly appears in keys
    assert "B" in graph and "C" in graph
# =============== END RequestForest tests ==============================

# ==================== START Failure and Cancelled tests =================
def test_cancelled_and_failure():
    c = Cancelled()
    assert isinstance(c, Cancelled)
    e = Exception("fail")
    f = Failure(e)
    assert isinstance(f, Failure)
    assert f.exception is e
# =============== END Failure and Cancelled tests ========================



def test_status():
    unfinished_request = RequestTemplate(
        identifier="ahsusu",
        stamp = Stamp(time=time.time(), step=0, identifier="Hello World"),
        parent=None,
        source_id="heghe",
        sink_id="heheuii",
        input=((), {}),
        output=None,
    )

    assert unfinished_request.status == "Open"

def test_status_2():
    completed_request = RequestTemplate(
        identifier="ahsusu",
        stamp=Stamp(time=time.time(), step=0, identifier="Hello World"),
        parent=None,
        source_id="heghe",
        sink_id="heheuii",
        input=((), {}),
        output="example output",
    )
    assert completed_request.status == "Completed"

def test_status_failure():
    failed_request = RequestTemplate(
        identifier="ahsusu",
        stamp=Stamp(time=time.time(), step=0, identifier="Hello World"),
        parent=None,
        source_id="heghe",
        sink_id="heheuii",
        input=((), {}),
        output=Failure(Exception("example")),
    )

    assert failed_request.status == "Failed"



from ._call import call
from .batch import call_batch
from .broadcast_ import broadcast
from .interactive import local_chat

__all__ = [
    "call",
    "call_batch",
    "broadcast",
    "local_chat",
]



from __future__ import annotations

import asyncio
import warnings
from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Generator,
    Generic,
    Literal,
    ParamSpec,
    Set,
    Type,
    TypeVar,
)

from railtracks.built_nodes.concrete.response import LLMResponse
from railtracks.exceptions import LLMError, NodeCreationError
from railtracks.interaction._call import call
from railtracks.llm import (
    AssistantMessage,
    Message,
    MessageHistory,
    ModelBase,
    ToolCall,
    ToolMessage,
    ToolResponse,
    UserMessage,
)
from railtracks.llm.content import Content
from railtracks.llm.message import Role
from railtracks.llm.providers import ModelProvider
from railtracks.llm.response import Response
from railtracks.nodes.nodes import Node
from railtracks.validation.node_creation.validation import check_connected_nodes
from railtracks.validation.node_invocation.validation import check_max_tool_calls

from ._llm_base import LLMBase

_T = TypeVar("_T")
_P = ParamSpec("_P")
_TStream = TypeVar("_TStream", Literal[True], Literal[False])
_TCollectedOutput = TypeVar("_TCollectedOutput", bound=LLMResponse)

_TContent = TypeVar("_TContent", bound=Content)


class OutputLessToolCallLLMBase(
    LLMBase[_T, _TCollectedOutput, _TStream],
    ABC,
    Generic[_T, _TCollectedOutput, _TStream],
):
    """A base class that is a node which contains
     an LLm that can make tool calls. The tool calls will be returned
    as calls or if there is a response, the response will be returned as an output"""

    def __init_subclass__(cls):
        super().__init_subclass__()
        # 3. Check if the tool_nodes is not empty, special case for ToolCallLLM
        # We will not check for abstract classes
        has_abstract_methods = any(
            getattr(getattr(cls, name, None), "__isabstractmethod__", False)
            for name in dir(cls)
        )
        if not has_abstract_methods:
            if "tool_nodes" in cls.__dict__ and not has_abstract_methods:
                method = cls.__dict__["tool_nodes"]
                try:
                    # Try to call the method as a classmethod (typical case)
                    node_set = method.__func__(cls)
                except AttributeError:
                    # If that fails, call it as an instance method (for easy_wrapper init)
                    dummy = object.__new__(cls)
                    node_set = method(dummy)
                # Validate that the returned node_set is correct and contains only Node/function instances
                check_connected_nodes(node_set, Node)

    @classmethod
    def streaming_blacklist(cls):
        return {
            ModelProvider.ANTHROPIC,
            ModelProvider.AZUREAI,
            ModelProvider.GEMINI,
            ModelProvider.OLLAMA,
            ModelProvider.HUGGINGFACE,
        }

    def __init__(
        self,
        user_input: MessageHistory | UserMessage | str | list[Message],
        llm: ModelBase[_TStream] | None = None,
        max_tool_calls: int | None = None,
    ):
        super().__init__(llm=llm, user_input=user_input)
        model = self.get_llm()
        # we only support Openai for streaming calls atm.
        if (
            model is not None
            and model.stream
            and model.model_provider() in self.streaming_blacklist()
        ):
            raise NodeCreationError(
                f"Currently we do not allow streaming with {model.model_provider()} (specifically for tool calling)",
                notes=[
                    "Create a new issue on the railtracks repo or switch to openai's models"
                ],
            )
        # Set max_tool_calls for non easy usage wrappers
        if not hasattr(self, "max_tool_calls"):
            # Check max_tool_calls (including warning for None)
            check_max_tool_calls(max_tool_calls)
            self.max_tool_calls = max_tool_calls

        # Warn user that two max_tool_calls are set and we will use the parameter
        else:
            if max_tool_calls is not None:
                warnings.warn(
                    "You have provided max_tool_calls as a parameter and as a class variable. We will use the parameter."
                )
                check_max_tool_calls(max_tool_calls)
                self.max_tool_calls = max_tool_calls
            else:
                check_max_tool_calls(self.max_tool_calls)

    @classmethod
    def name(cls) -> str:
        return "Tool Call LLM"

    @classmethod
    @abstractmethod
    def tool_nodes(cls) -> Set[Type[Node]]: ...

    def create_node(self, tool_name: str, arguments: Dict[str, Any]) -> Node:
        """
        A function which creates a new instance of a node Class from a tool name and arguments.

        This function may be overwritten to fit the needs of the given node as needed.
        """
        node = [x for x in self.tool_nodes() if x.tool_info().name == tool_name]
        if node == []:
            raise LLMError(
                reason=f" Error creating a node from tool {tool_name}. The tool_name given by the LLM doesn't match any of the tool names in the connected nodes.",
                message_history=self.message_hist,
            )
        if len(node) > 1:
            raise NodeCreationError(
                message=f"Tool {tool_name} has multiple nodes, this is not allowed. Current Node include {[x.tool_info().name for x in self.tool_nodes()]}",
                notes=["Please check the tool names in the connected nodes."],
            )
        return node[0].prepare_tool(arguments)

    def get_node_from_name(self, tool_name: str):
        """
        Gets the node attached to the node of the given name. If no node exists or there are multiple matches, it will raise an exception.
        """
        node = [x for x in self.tool_nodes() if x.tool_info().name == tool_name]
        if node == []:
            raise LLMError(
                reason=f"Error creating a node from tool {tool_name}. The tool_name given by the LLM doesn't match any of the tool names in the connected nodes.",
                message_history=self.message_hist,
            )
        if len(node) > 1:
            raise NodeCreationError(
                message=f"Tool {tool_name} has multiple nodes, this is not allowed. Current Node include {[x.tool_info().name for x in self.tool_nodes()]}",
                notes=["Please check the tool names in the connected nodes."],
            )

        return node[0]

    async def run_node_from_tool(self, tool_name: str, arguments: dict[str, Any]):
        node = self.get_node_from_name(tool_name)

        return await call(node.prepare_tool, **arguments)

    @classmethod
    def tools(cls):
        return [x.tool_info() for x in cls.tool_nodes()]

    async def _call_tools(self, tool_calls: list[ToolCall]) -> list[ToolMessage]:
        contracts = []

        for t_c in tool_calls:
            contract = self.run_node_from_tool(t_c.name, t_c.arguments)
            contracts.append(contract)

        tool_responses = await asyncio.gather(*contracts, return_exceptions=True)
        tool_responses = [
            (
                x
                if not isinstance(x, Exception)
                else f"There was an error running the tool: \n Exception message: {x} "
            )
            for x in tool_responses
        ]
        tool_ids = [x.identifier for x in tool_calls]
        tool_names = [x.name for x in tool_calls]

        tool_messages = []

        for r_id, r_name, resp in zip(
            tool_ids,
            tool_names,
            tool_responses,
        ):
            tool_messages.append(
                ToolMessage(
                    ToolResponse(identifier=r_id, result=str(resp), name=r_name)
                )
            )

        return tool_messages

    async def _handle_response(
        self,
        message: Message[_TContent, Literal[Role.assistant]],
        allowed_tool_calls: int | None,
    ):
        # if the returned item is a list then it is a list of tool calls
        if isinstance(message.content, list):
            assert all(isinstance(x, ToolCall) for x in message.content)

            tool_calls = message.content
            if allowed_tool_calls is not None and len(tool_calls) > allowed_tool_calls:
                tool_calls = tool_calls[:allowed_tool_calls]

            # append the requested tool calls assistant message, once the tool calls have been verified and truncated (if needed)
            self.message_hist.append(AssistantMessage(content=tool_calls))

            tool_messages = await self._call_tools(tool_calls)
            for t_m in tool_messages:
                self.message_hist.append(t_m)

            return True, None
        else:
            # this means the tool call is finished
            self.message_hist.append(message)
            return False, message

    def _get_allowed_tool_calls(self) -> int | None:
        current_tool_calls = len(
            [m for m in self.message_hist if isinstance(m, ToolMessage)]
        )
        allowed_tool_calls = (
            self.max_tool_calls - current_tool_calls
            if self.max_tool_calls is not None
            else None
        )
        return allowed_tool_calls


class OutputLessToolCallLLM(
    OutputLessToolCallLLMBase[_TCollectedOutput, _TCollectedOutput, Literal[False]],
    ABC,
    Generic[_TCollectedOutput],
):
    async def _on_max_tool_calls_exceeded(self):
        """force a final response. This function will add it to the message history and return the message. This function shoudld only be called on non streaming llms."""
        response = await asyncio.to_thread(
            self.llm_model.chat,
            self.message_hist,
        )

        if not response.message.role == Role.assistant:
            raise LLMError(
                reason=f"The LLM returned an unexpected message type. Expected AssistantMessage but got {type(response.message)}",
                message_history=self.message_hist,
            )

        return response.message

    async def _handle_tool_calls(self) -> tuple[bool, Message | None]:
        """
        Handles the execution of tool calls for the node, including LLM interaction and message history updates.

        This method:
        - Checks if the maximum number of tool calls has been reached and triggers a final response if so.
        - Interacts with the LLM to get a tool call request or final answers.
        - Executes a tool call and appends the results to the message history.
        - Handles malformed LLM responses and raises errors as needed.

        Returns:
            bool: True if more tool calls are expected (the tool call loop should continue),
                  False if the tool call process is finished and a final answer is available.

        Raises:
            LLMError: If the LLM returns an unexpected message type or the message is malformed.
        """
        allowed_tool_calls = self._get_allowed_tool_calls()

        if allowed_tool_calls is not None and allowed_tool_calls <= 0:
            message = await self._on_max_tool_calls_exceeded()
            return False, message

        # collect the response from the llm model
        response = await asyncio.to_thread(
            self.llm_model.chat_with_tools, self.message_hist, tools=self.tools()
        )

        if not response.message.role == Role.assistant:
            raise LLMError(
                reason=f"The LLM returned an unexpected message type. Expected AssistantMessage but got {type(response.message)}",
                message_history=self.message_hist,
            )

        return await self._handle_response(response.message, allowed_tool_calls)

    async def invoke(self):
        message = None
        while True:
            still_tool_calls, message = await self._handle_tool_calls()
            if not still_tool_calls:
                break

        return self.return_output(message)


class StreamingOutputLessToolCallLLM(
    OutputLessToolCallLLMBase[
        Generator[str | _TCollectedOutput, None, _TCollectedOutput],
        _TCollectedOutput,
        Literal[True],
    ],
    ABC,
    Generic[_TCollectedOutput],
):
    async def _handle_tool_calls(self):
        returned_mess = await asyncio.to_thread(
            self.llm_model.chat_with_tools, self.message_hist, tools=self.tools()
        )

        first_item = next(returned_mess)
        if isinstance(first_item, str):

            def gen_wrapper():
                yield first_item
                # yield the rest of the items
                for chunk in returned_mess:
                    if isinstance(chunk, str):
                        yield chunk
                    elif isinstance(chunk, Response):
                        if chunk.message.role != Role.assistant:
                            raise LLMError(
                                reason="ModelLLM returned an unexpected message type.",
                                message_history=self.message_hist,
                            )
                        self.message_hist.append(chunk.message)
                        response = self.return_output(chunk.message)
                        yield response
                        return response

                raise LLMError(
                    "Badly formatted response from the LLM",
                    message_history=self.message_hist,
                )

            return gen_wrapper()

        if isinstance(first_item, Response):
            assert first_item.message.role == Role.assistant

            if len(first_item.message.tool_calls) > 0:
                is_tool, _ = await self._handle_response(
                    first_item.message, self._get_allowed_tool_calls()
                )

                if not is_tool:
                    raise LLMError(
                        "Message returned did not contain tool calls and it should have.",
                        message_history=self.message_hist,
                    )
            else:
                raise LLMError(
                    "Message returned did not contain tool calls and it should have.",
                    message_history=self.message_hist,
                )

    async def invoke(self):
        """Makes a call containing the inputted message and system prompt to the llm model and returns the response
        Returns:
            (TerminalLLM.Output): The response message from the llm model
        """

        while True:
            result = await self._handle_tool_calls()
            if result is not None:
                return result



import pytest

from typing import Callable, Any, Optional

from railtracks.vector_stores.vector_store_base import (
    Chunk,
    VectorStore,
    OneOrMany,
)


class MockVectorStore(VectorStore):
    """
    A concrete in-memory mock of VectorStore for testing and development.

    You can optionally pass in custom functions to override behavior of any
    method (upsert, search, fetch, delete, count). If not provided, default
    in-memory behavior is used.

    Example:
        store = MockVectorStore(
            collection_name="test",
            embedding_function=lambda x: [[0.0]],
            search_func=lambda query, top_k, **_: [...]
        )
    """

    def __init__(
        self,
        collection_name: str = "mock",
        embedding_function: Callable[[list[str]], list[list[float]]] = lambda x: [[0.0] for _ in x],
        *,
        upsert_func: Optional[Callable] = None,
        search_func: Optional[Callable] = None,
        fetch_func: Optional[Callable] = None,
        delete_func: Optional[Callable] = None,
        count_func: Optional[Callable] = None,
    ):
        super().__init__(collection_name, embedding_function)

        # Optional user-provided overrides
        self._custom_upsert = upsert_func
        self._custom_search = search_func
        self._custom_fetch = fetch_func
        self._custom_delete = delete_func
        self._custom_count = count_func


    # -----------------------------
    # Utility helpers
    # -----------------------------
    def _ensure_list(self, x: OneOrMany[Any]) -> list[Any]:
        return x if isinstance(x, list) else [x]


    def _make_chunk(self, value: Chunk | str):
        if isinstance(value, Chunk):
            return value
        return Chunk(content=value)

    # -----------------------------
    # VectorStore interface
    # -----------------------------
    def upsert(
        self,
        content: OneOrMany[Chunk] | OneOrMany[str],
    ):
        """Insert or update vectors into the mock store."""
        if self._custom_upsert:
            return self._custom_upsert(content)

        raise NotImplementedError()



    def fetch(self, ids: OneOrMany[str]):
        """Fetch chunks from the mock store by id."""
        if self._custom_fetch:
            return self._custom_fetch(ids)

        raise NotImplementedError()
    

    def search(
        self,
        query: OneOrMany[Chunk] | OneOrMany[str],
        top_k: int = 10,
        where: Optional[dict[str, Any]] = None,
        include: Optional[list[str]] = None,
    ):
        """
        Very dumb default: returns the *first top_k items* in the store,
        ignoring the actual query. Intended for tests unless overridden.
        """
        if self._custom_search:
            return self._custom_search(query, top_k, where=where, include=include)

        raise NotImplementedError()

    def delete(self, ids: OneOrMany[str], where: Optional[dict[str, Any]] = None):
        """Deletes items by id. Ignores metadata unless overridden."""
        if self._custom_delete:
            return self._custom_delete(ids, where)

        raise NotImplementedError()

    def count(self) -> int:
        """Returns number of stored vectors."""
        if self._custom_count:
            return self._custom_count()
        
        raise NotImplementedError()

@pytest.fixture
def mock_vector_store():
    return MockVectorStore()



from typing import List

import railtracks as rt
from railtracks.rag.rag_core import RAG, RAGConfig, SearchResult


def rag_node(
    documents: List[str],
    embed_model="text-embedding-3-small",
    token_count_model="gpt-4o",
    chunk_size=1000,
    chunk_overlap=200,
):
    """
    Creates a rag node that allows you to vector the search the provided documents.

    Args:
        documents (list): List of documents to process. Each document is a string of raw text.
        embed_model (str): Model name for embedding service.
        token_count_model (str): Model name for token counting.
        chunk_size (int): Size of each text chunk.
        chunk_overlap (int): Overlap between chunks.

    Returns:
        A node to be invoked upon request.

    """

    rag_core = RAG(
        docs=documents,
        config=RAGConfig(
            embedding={"model": embed_model},
            store={},
            chunking={
                "chunk_size": chunk_size,
                "chunk_overlap": chunk_overlap,
                "model": token_count_model,
            },
        ),
    )
    rag_core.embed_all()

    def query(query: str, top_k: int = 1) -> SearchResult:
        result = rag_core.search(query, top_k=top_k)
        return result

    return rt.function_node(query)



import pytest
from pathlib import Path
import shutil


@pytest.fixture(scope="session", autouse=True)
def global_teardown():
    # Setup code (before tests run)
    yield
    # Teardown code (after all tests run)
    railtracks_dir = Path(".railtracks")
    if railtracks_dir.exists() and railtracks_dir.is_dir():
        shutil.rmtree(railtracks_dir)
        print("Cleaned up .railtracks directory after tests.")


import asyncio

import railtracks as rt

import random
from datetime import datetime, timedelta

@rt.function_node
async def search_flights():
    await asyncio.sleep(1)
    cities = ["New York", "Los Angeles", "Chicago", "Houston"]
    airlines = [
        "American Airlines",
        "Delta Airlines",
        "United Airlines",
        "Southwest Airlines",
        "British Airways",
        "Air France",
    ]
    flights = []

    for i in range(150):  # Generate 50 flights
        dep_city = random.choice(cities)
        arr_city = random.choice(
            [city for city in cities if city != dep_city]
        )  # Ensure arrival city is different
        departure_time = datetime.now() + timedelta(
            days=random.randint(1, 30), hours=random.randint(0, 23)
        )
        arrival_time = departure_time + timedelta(
            hours=random.randint(1, 6)
        )  # Flight duration between 1-6 hours
        flight = {
            "flight_number": f"{random.choice(['AA', 'DL', 'UA', 'SW', 'BA', 'AF'])}{random.randint(100, 999)}",
            "airline": random.choice(airlines),
            "departure_city": dep_city,
            "arrival_city": arr_city,
            "departure_time": departure_time.strftime("%Y-%m-%dT%H:%M:%S"),
            "arrival_time": arrival_time.strftime("%Y-%m-%dT%H:%M:%S"),
            "price": round(
                random.uniform(100, 1000), 2
            ),  # Price between $100 and $1000
        }
        flights.append(flight)

    return flights


@rt.function_node
def filter_flights(flights: list, departure_city: str, arrival_city: str):
    filtered_flights = [
        flight
        for flight in flights
        if flight["departure_city"] == departure_city
        and flight["arrival_city"] == arrival_city
    ]
    if not filtered_flights:
        raise ValueError(
            f"No outbound flights found from {departure_city} to {arrival_city}."
        )
    return random.choice(filtered_flights)


@rt.function_node
async def pick_flight(flights: list, departure_city: str, arrival_city: str):
    contracts = asyncio.gather(
        rt.call(filter_flights, flights, departure_city, arrival_city),
        rt.call(filter_flights, flights, arrival_city, departure_city),
    )
    outbound, inbound = await contracts

    return {"outbound_flight": outbound, "inbound_flight": inbound}


@rt.function_node
async def book_flight(flight):
    await asyncio.sleep(1)
    return f"Flight {flight['flight_number']} booked from {flight['departure_city']} to {flight['arrival_city']}."


@rt.function_node
async def planner(current_city: str, destination_city: str):
    flights = await rt.call(search_flights)

    flight_plan = await rt.call(pick_flight, flights, current_city, destination_city)

    outbound_booking = await rt.call(book_flight, flight_plan["outbound_flight"])
    inbound_booking = await rt.call(book_flight, flight_plan["inbound_flight"])
    return {"outbound_booking": outbound_booking, "inbound_booking": inbound_booking}


TLLMNode = rt.agent_node(
    system_message="You are excellent chooser of random items from a list. You never make mistakes and have a god like ability to accomplish tasks.",
)

@rt.function_node
async def planner_with_llm(llm: rt.llm.ModelBase):
    cities = []
    available_cities = ["New York", "Chicago", "Los Angeles", "Houston"]
    while True:

        result = await rt.call(
            TLLMNode, 
            user_input=f"Choose a Random city from {available_cities} and give me the index of the city only. Nothing else.",
            llm=llm,
        )  # we will make the mock_llm return a random integer between 0 and 3 to simulate the random choice

        city = available_cities[int(result.text)]
        if city in available_cities:
            available_cities.remove(city)
            cities.append(city)

        if len(cities) == 2:
            break

    return await rt.call(planner, *cities)


@pytest.fixture
def planner_node():
    return planner


@pytest.fixture
def planner_with_llm_node():
    return planner_with_llm


import json


@pytest.fixture
def json_state_schema():
    json_path = Path(__file__).parent / "state_schema.json"
    with open(json_path, "r") as f:
        return json.load(f)



"""Parameter class for object types."""

from typing import Any, Dict, Optional

from ._base import Parameter, ParameterType


class ObjectParameter(Parameter):
    """Parameter representing an object type."""

    param_type: ParameterType = ParameterType.OBJECT

    def __init__(
        self,
        name: str,
        properties: list[Parameter],
        description: Optional[str] = None,
        required: bool = True,
        additional_properties: bool = False,
        default: Any = None,
    ):
        """Initialize an ObjectParameter instance.

        Args:
            name (str): Name of the parameter.
            properties (list[Parameter]): List of Parameter instances describing object properties.
            description (Optional[str]): Description of the parameter.
            required (bool): Whether the parameter is required.
            additional_properties (bool): Whether additional properties are allowed.
            default (Any): Default value for the parameter.
        """
        super().__init__(name, description, required, default)
        self.properties = properties
        self.additional_properties = additional_properties

    def to_json_schema(self) -> Dict[str, Any]:
        """Convert the object parameter to a JSON schema representation."""
        schema = {
            "type": "object",
            "properties": {},
            "additionalProperties": self.additional_properties,
        }

        if self.description:
            schema["description"] = self.description

        required_props = []
        for prop in self.properties:
            schema["properties"][prop.name] = prop.to_json_schema()
            if prop.required:
                required_props.append(prop.name)

        if required_props:
            schema["required"] = required_props

        if self.default is not None:
            schema["default"] = self.default

        if self.enum:
            schema["enum"] = self.enum

        return schema

    def __repr__(self) -> str:
        """Return a string representation of the ObjectParameter."""
        return (
            f"ObjectParameter(name={self.name!r}, properties={self.properties!r}, "
            f"description={self.description!r}, required={self.required!r}, "
            f"additional_properties={self.additional_properties!r}, default={self.default!r})"
        )



from __future__ import annotations

import json
from typing import Any, List, Tuple, TypeVar

from railtracks.utils.profiling import Stamp, StampManager
from railtracks.utils.serialization.graph import Edge, Vertex

from .node import NodeForest
from .request import RequestForest
from .serialize import RTJSONEncoder
from .utils import create_sub_state_info

_TOutput = TypeVar("_TOutput")


# TODO add some the logic for an optional architecture requirement.
class ExecutionInfo:
    """
    A class that contains the full details of the state of a run at any given point in time.

    The class is designed to be used as a snapshot of state that can be used to display the state of the run, or to
    create a graphical representation of the system.
    """

    def __init__(
        self,
        request_forest: RequestForest,
        node_forest: NodeForest,
        stamper: StampManager,
    ):
        self.request_forest = request_forest
        self.node_forest = node_forest
        self.stamper = stamper

    @classmethod
    def default(cls) -> ExecutionInfo:
        """Creates a new "empty" instance of the ExecutionInfo class with the default values."""
        return cls.create_new()

    @classmethod
    def create_new(
        cls,
    ) -> ExecutionInfo:
        """
        Creates a new empty instance of state variables with the provided executor configuration.

        """
        request_heap = RequestForest()
        node_heap = NodeForest()
        stamper = StampManager()

        return ExecutionInfo(
            request_forest=request_heap,
            node_forest=node_heap,
            stamper=stamper,
        )

    @property
    def answer(self):
        """Convenience method to access the answer of the run."""
        return self.request_forest.answer

    @property
    def all_stamps(self) -> List[Stamp]:
        """Convenience method to access all the stamps of the run."""
        return self.stamper.all_stamps

    @property
    def name(self):
        """
        Gets the name of the graph by pulling the name of the insertion request. It will raise a ValueError if the insertion
        request is not present or there are multiple insertion requests.
        """
        insertion_requests = self.insertion_requests

        if len(insertion_requests) >= 2:
            raise ValueError(
                "You cannot get the name of a graph with multiple insertion requests"
            )

        if len(insertion_requests) == 0:
            raise ValueError(
                "You cannot get the name of a graph with no insertion requests"
            )

        i_r = insertion_requests[0]

        return self.node_forest.get_node_type(i_r.sink_id).name()

    @property
    def insertion_requests(self):
        """A convenience method to access all the insertion requests of the run."""
        return self.request_forest.insertion_request

    def _get_info(self, ids: List[str] | str | None = None) -> ExecutionInfo:
        """
        Gets a subset of the current state based on the provided node ids. It will contain all the children of the provided node ids

        Note: If no ids are provided, the full state is returned.

        Args:
            ids (List[str] | str | None): A list of node ids to filter the state by. If None, the full state is returned.

        Returns:
            ExecutionInfo: A new instance of ExecutionInfo containing only the children of the provided ids.

        """
        if ids is None:
            return self
        else:
            # firstly lets
            if isinstance(ids, str):
                ids = [ids]

            # we need to quickly check to make sure these ids are valid
            for identifier in ids:
                if identifier not in self.request_forest:
                    raise ValueError(
                        f"Identifier '{identifier}' not found in the current state."
                    )

            new_node_forest, new_request_forest = create_sub_state_info(
                self.node_forest.heap(),
                self.request_forest.heap(),
                ids,
            )
            return ExecutionInfo(
                node_forest=new_node_forest,
                request_forest=new_request_forest,
                stamper=self.stamper,
            )

    def _to_graph(self) -> Tuple[List[Vertex], List[Edge]]:
        """
        Converts the current state into its graph representation.

        Returns:
            List[Node]: An iterable of nodes in the graph.
            List[Edge]: An iterable of edges in the graph.
        """
        return self.node_forest.to_vertices(), self.request_forest.to_edges()

    def graph_serialization(self) -> dict[str, Any]:
        """
                Creates a string (JSON) representation of this info object designed to be used to construct a graph for this
                info object.

                Some important notes about its structure are outlined below:
                - The `nodes` key contains a list of all the nodes in the graph, represented as `Vertex` objects.
                - The `edges` key contains a list of all the edges in the graph, represented as `Edge` objects.
                - The `stamps` key contains an ease of use list of all the stamps associated with the run, represented as `Stamp` objects.

                - The "nodes" and "requests" key will be outlined with normal graph details like connections and identifiers in addition to a loose details object.
                - However, both will carry an addition param called "stamp" which is a timestamp style object.
                - They also will carry a "parent" param which is a recursive structure that allows you to traverse the graph in time.


        ```
        """
        parent_nodes = [x.identifier for x in self.insertion_requests]

        infos = [self._get_info(parent_node) for parent_node in parent_nodes]

        runs = []

        for info, parent_node_id in zip(infos, parent_nodes):
            insertion_requests = info.request_forest.insertion_request

            assert len(insertion_requests) == 1
            parent_request = insertion_requests[0]

            all_parents = parent_request.get_all_parents()

            start_time = all_parents[-1].stamp.time

            assert len([x for x in all_parents if x.status == "Completed"]) <= 1
            end_time = None
            for req in all_parents:
                if req.status in ["Completed", "Failed"]:
                    end_time = req.stamp.time
                    break

            entry = {
                "name": info.name,
                "run_id": parent_node_id,
                "nodes": info.node_forest.to_vertices(),
                "status": parent_request.status,
                "edges": info.request_forest.to_edges(),
                "steps": _get_stamps_from_forests(
                    info.node_forest, info.request_forest
                ),
                "start_time": start_time,
                "end_time": end_time,
            }
            runs.append(entry)

        return json.loads(
            json.dumps(
                runs,
                cls=RTJSONEncoder,
            )
        )


def _get_stamps_from_forests(
    node_forest: NodeForest,
    request_forest: RequestForest,
):
    node_stamps = {n.stamp for n in node_forest.full_data()}
    request_stamps = {r.stamp for r in request_forest.full_data()}

    result = sorted(node_stamps.union(request_stamps))

    return result



from pydantic import BaseModel
import pytest
import railtracks as rt



@pytest.fixture
def schema():
    class DummySchema(BaseModel):
        value: int

    return DummySchema

@pytest.fixture
def mock_tool():
    @rt.function_node
    def dummy_tool(param1: int, param2: str) -> str:
        return f"param1: {param1}, param2: {param2}"
    
    return dummy_tool.node_type


import asyncio
import time

import railtracks as rt
from railtracks.nodes.nodes import Node

import pytest
import subprocess
import sys

from railtracks.rt_mcp.main import MCPHttpParams, MCPStdioParams


@pytest.fixture(scope="session", autouse=True)
def install_mcp_server_time():
    """Install mcp_server_time and ensure mcp dependency is available.
    
    This fixture ensures that both mcp and mcp_server_time are properly installed,
    which is particularly important on Windows where dependency resolution
    may not work correctly when packages are installed separately.
    """
    try:
        # Install both packages together to ensure proper dependency resolution
        subprocess.check_call([
            sys.executable, "-m", "pip", "install", 
            "mcp>=1.9.0", "mcp_server_time"
        ])
        
        # Verify that the critical modules can be imported
        # This helps catch environment/path issues early
        try:
            import mcp.server  # This is what mcp_server_time needs
            import mcp_server_time
        except ImportError as e:
            pytest.fail(f"MCP packages installed but cannot be imported: {e}")
            
    except subprocess.CalledProcessError as e:
        pytest.fail(f"Failed to install required MCP packages: {e}")


def test_from_mcp_server_basic():
    time_server = rt.connect_mcp(
        MCPStdioParams(
            command=sys.executable,
            args=["-m", "mcp_server_time", "--local-timezone=America/Vancouver"],
        )
    )
    assert len(time_server.tools) == 2
    assert all(issubclass(tool, Node) for tool in time_server.tools)



class MockClient:
    def __init__(self, delay=1):
        self.delay = delay

    async def call_tool(self, tool_name, kwargs):
        await asyncio.sleep(self.delay)
        return f"done {tool_name}"

    async def list_tools(self):
        Tool = type("Tool", (), {
            "name": "tool1",
            "description": "Mock tool 1",
            "inputSchema": {
                "type": "object",
                "properties": {},
                "required": []
            }
        })
        Tool2 = type("Tool", (), {
            "name": "tool2",
            "description": "Mock tool 2",
            "inputSchema": {
                "type": "object",
                "properties": {},
                "required": []
            }
        })
        return type("ToolResponse", (), {"tools": [Tool, Tool2]})()


@pytest.mark.asyncio
async def test_parallel_mcp_servers():
    client = MockClient()
    client2 = MockClient()
    server1 = rt.connect_mcp(MCPHttpParams(url=""), client).tools[0]
    server2 = rt.connect_mcp(MCPHttpParams(url=""), client2).tools[1]

    start = time.perf_counter()
    results = await asyncio.gather(rt.call(server1), rt.call(server2))
    elapsed = time.perf_counter() - start

    assert all("done" in r for r in results)
    assert elapsed < 2



# =================== START fixtures for node.py ====================
import pytest
import uuid
from railtracks.state.node import (
    LinkedNode, NodeCopyError,
)
from railtracks.utils.profiling import Stamp

# =============== START LinkedNode basic/tests =====================

def test_linkednode_props(dummy_node_factory, linked_node_factory):
    identifier = str(uuid.uuid4())
    stamp = Stamp(123, 1, "test")
    node = dummy_node_factory(uuid=identifier, details={'x': 42})
    parent = None
    ln = LinkedNode(identifier=identifier, _node=node, stamp=stamp, parent=parent)

    # The .node property should return a copy, not original, and error if copying fails
    node2 = ln.node
    assert isinstance(node2, type(node))
    assert node2._copied is True  # Check it's indeed a "copied" node

def test_linkednode_to_vertex_includes_hierarchy(dummy_node_factory, linked_node_factory):
    base_id = str(uuid.uuid4())
    parent_stamp = Stamp(1, 0, "init")
    parent_ln = linked_node_factory(base_id, parent_stamp, parent=None, pretty='parent', details={'a': 1})
    child_stamp = Stamp(1, 1, "child")
    child_ln = linked_node_factory(base_id, child_stamp, parent=parent_ln, pretty='child', details={'b': 2})

    vertex = child_ln.to_vertex()
    assert vertex.identifier == child_ln.identifier
    assert vertex.name == child_ln.node.name()
    assert vertex.stamp == child_ln.stamp
    assert vertex.parent.identifier == parent_ln.identifier
    assert vertex.node_type == parent_ln.node.type()

def test_linkednode_nodecopyerror_raised_on_safe_copy(monkeypatch, linked_node_factory):
    class UnsafeNode:
        def safe_copy(self):
            raise RuntimeError("fail copy")
        @property
        def pretty_name(self):
            return 'unsafe'

    ln = LinkedNode(
        identifier='id', _node=UnsafeNode(), stamp=Stamp(1, 1, "fail"), parent=None
    )
    with pytest.raises(NodeCopyError):
        _ = ln.node
# =============== END LinkedNode basic/tests =======================

# =============== START NodeForest heap & access ===================
def test_nodeforest_update_and_getitem(node_forest, dummy_node_factory):
    a_id = str(uuid.uuid4())
    stamp = Stamp(1, 2, "s1")
    node = dummy_node_factory(uuid=a_id, pretty='ntype')
    node_forest.update(node, stamp)
    # __getitem__ returns LinkedNode
    linked = node_forest[a_id]
    assert isinstance(linked, LinkedNode)
    assert linked.identifier == a_id
    # Returns a copied node
    n_copied = linked.node
    assert n_copied.uuid == node.uuid
    assert getattr(n_copied, "_copied", False)

def test_nodeforest_multiple_updates_chain(node_forest, dummy_node_factory):
    a_id = str(uuid.uuid4())
    # Add first
    node1 = dummy_node_factory(uuid=a_id, pretty='first')
    stamp1 = Stamp(1, 1, "first")
    node_forest.update(node1, stamp1)
    # Add again (should chain parent)
    node2 = dummy_node_factory(uuid=a_id, pretty='second')
    stamp2 = Stamp(1, 2, "second")
    node_forest.update(node2, stamp2)
    ln2 = node_forest[a_id]
    assert ln2.stamp == stamp2
    assert ln2.parent.stamp == stamp1
    assert ln2.parent.parent is None

def test_nodeforest_id_type_mapping(node_forest, dummy_node_factory):
    a_id = str(uuid.uuid4())
    node = dummy_node_factory(uuid=a_id, pretty='ABrandNewType')
    stamp = Stamp(1, 9, "mapping")
    node_forest.update(node, stamp)
    typ = node_forest.get_node_type(a_id)
    assert typ == type(node)

# =============== END NodeForest heap & access =====================

# =============== START to_vertices & conversions ==================
def test_nodeforest_to_vertices_returns_all(node_forest, dummy_node_factory):
    id1 = str(uuid.uuid4())
    id2 = str(uuid.uuid4())
    stamp1 = Stamp(1, 1, "a")
    stamp2 = Stamp(2, 2, "b")
    node_forest.update(dummy_node_factory(uuid=id1, pretty='nt1'), stamp1)
    node_forest.update(dummy_node_factory(uuid=id2, pretty='nt2'), stamp2)
    vertices = node_forest.to_vertices()
    assert len(vertices) == 2
    ids = {v.identifier for v in vertices}
    assert id1 in ids and id2 in ids

def test_nodeforest_to_vertices_hierarchical(node_forest, dummy_node_factory):
    id1 = str(uuid.uuid4())
    stamp1 = Stamp(1, 1, "root")
    node_forest.update(dummy_node_factory(uuid=id1, pretty='nt1'), stamp1)
    stamp2 = Stamp(1, 2, "child")
    node_forest.update(dummy_node_factory(uuid=id1, pretty='nt1'), stamp2)
    vertices = node_forest.to_vertices()
    # The newest should contain the parent as .parent
    this_vertex = [v for v in vertices if v.stamp == stamp2][0]
    assert this_vertex.parent is not None
    assert this_vertex.parent.identifier == id1
# =============== END to_vertices & conversions ====================


import pytest
import railtracks as rt

from railtracks.nodes.nodes import Node
from railtracks.prebuilt import rag_node




@pytest.mark.asyncio
async def test_node_search_question(get_docs):

    docs = get_docs
    node = rag_node(
        documents=docs,
    )

    query = "What is the color of watermelon?"
    result = await rt.call(node, query, 2)
    print(query)
    print(result[0].record.text)
    print(result[1].record.text)
    assert isinstance(result, list)
    assert len(result) > 0
    # doc[2] should contain the watermelon description
    assert result[0].record.text == docs[2]

@pytest.mark.asyncio
async def test_node_search_confirmation(get_docs):
    docs = get_docs
    node = rag_node(
        documents=docs,
    )

    query = "Pear is yellow"
    result = await rt.call(node, query, 2)
    print(query)
    print(result[0].record.text)
    print(result[1].record.text)
    assert isinstance(result, list)
    assert len(result) > 0
    # doc[2] should contain the watermelon description
    assert result[0].record.text == docs[1]



import asyncio
import threading
from contextlib import AsyncExitStack
from datetime import timedelta
from typing import Any

from mcp import ClientSession, StdioServerParameters
from mcp.client.sse import sse_client
from mcp.client.stdio import stdio_client
from mcp.client.streamable_http import streamablehttp_client
from pydantic import BaseModel
from typing_extensions import Self, Type

from railtracks.llm import Tool
from railtracks.nodes.nodes import Node


class MCPStdioParams(StdioServerParameters):
    """
    Configuration parameters for STDIO-based MCP server connections.

    Extends the standard StdioServerParameters with a timeout field.

    Attributes:
        timeout: Maximum time to wait for operations (default: 30 seconds)
    """

    timeout: timedelta = timedelta(seconds=30)

    def as_stdio_params(self) -> StdioServerParameters:
        """
        Convert to standard StdioServerParameters, excluding the timeout field.

        Returns:
            StdioServerParameters without the timeout attribute
        """
        stdio_kwargs = self.dict(exclude={"timeout"})
        return StdioServerParameters(**stdio_kwargs)


class MCPHttpParams(BaseModel):
    """
    Configuration parameters for HTTP-based MCP server connections.

    Supports both SSE (Server-Sent Events) and streamable HTTP transports.
    The transport type is automatically determined based on the URL.

    Attributes:
        url: The MCP server URL (use /sse suffix for SSE transport)
        headers: Optional HTTP headers for authentication
        timeout: Connection timeout (default: 30 seconds)
        sse_read_timeout: SSE read timeout (default: 5 minutes)
        terminate_on_close: Whether to terminate connection on close (default: True)
    """

    url: str
    headers: dict[str, Any] | None = None
    timeout: timedelta = timedelta(seconds=30)
    sse_read_timeout: timedelta = timedelta(seconds=60 * 5)
    terminate_on_close: bool = True


class MCPAsyncClient:
    """
    Async client for communicating with an MCP server.

    Supports both STDIO and HTTP transports with streaming capabilities.
    Manages the connection lifecycle and provides methods for listing and calling tools.

    Attributes:
        config: Connection configuration (MCPStdioParams or MCPHttpParams)
        session: MCP client session (created automatically if not provided)
        exit_stack: Async context manager for resource cleanup

    Note:
        If a client session is provided, it will be used; otherwise, a new session
        will be created automatically based on the config type.
    """

    def __init__(
        self,
        config: MCPStdioParams | MCPHttpParams,
        client_session: ClientSession | None = None,
    ):
        self.config = config
        self.session = client_session
        self.exit_stack = AsyncExitStack()
        self._entered = False
        self._tools_cache = None

    async def connect(self):
        """
        Establish connection to the MCP server.

        Automatically selects the appropriate transport (STDIO or HTTP) based on config type.
        Creates and initializes the client session if one wasn't provided.

        Raises:
            ValueError: If config type is invalid
            Exception: If connection fails (will clean up resources automatically)
        """
        await self.exit_stack.__aenter__()
        self._entered = True
        try:
            if self.session is None:
                if isinstance(self.config, MCPStdioParams):
                    # STDIO transport for local MCP servers
                    stdio_transport = await self.exit_stack.enter_async_context(
                        stdio_client(self.config.as_stdio_params())
                    )
                    self.session = await self.exit_stack.enter_async_context(
                        ClientSession(*stdio_transport)
                    )
                    await self.session.initialize()
                elif isinstance(self.config, MCPHttpParams):
                    # HTTP transport (SSE or streamable HTTP)
                    await self._init_http()
                else:
                    raise ValueError(
                        "Invalid configuration type. Expected MCPStdioParams or MCPHttpParams."
                    )
        except Exception:
            await self.close()
            raise

    async def close(self):
        """Close the connection and clean up resources."""
        if self._entered:
            await self.exit_stack.aclose()
            self._entered = False

    async def list_tools(self):
        """
        List all tools available from the MCP server.

        Results are cached after the first call for efficiency.

        Returns:
            List of available tools
        """
        if self._tools_cache is not None:
            return self._tools_cache
        else:
            resp = await self.session.list_tools()
            self._tools_cache = resp.tools
        return self._tools_cache

    async def call_tool(self, tool_name: str, tool_args: dict):
        """
        Call a specific tool on the MCP server.

        Args:
            tool_name: Name of the tool to call
            tool_args: Arguments to pass to the tool

        Returns:
            Tool execution result
        """
        return await self.session.call_tool(tool_name, tool_args)

    async def _init_http(self):
        """
        Initialize HTTP-based connection (SSE or streamable HTTP).

        Automatically detects transport type based on URL:
        - URLs ending in /sse use Server-Sent Events transport
        - Other URLs use streamable HTTP transport

        Supports optional OAuth authentication via the 'auth' attribute.
        """
        # Determine transport type from URL
        if self.config.url.rstrip("/").endswith("/sse"):
            self.transport_type = "sse"
        else:
            self.transport_type = "streamable_http"

        # Create appropriate client based on transport type
        if self.transport_type == "sse":
            client = sse_client(
                url=self.config.url,
                headers=self.config.headers,
                timeout=self.config.timeout.total_seconds(),
                sse_read_timeout=self.config.sse_read_timeout.total_seconds(),
                auth=self.config.auth if hasattr(self.config, "auth") else None,
            )
        else:
            client = streamablehttp_client(
                url=self.config.url,
                headers=self.config.headers,
                timeout=self.config.timeout.total_seconds(),
                sse_read_timeout=self.config.sse_read_timeout.total_seconds(),
                terminate_on_close=self.config.terminate_on_close,
                auth=self.config.auth if hasattr(self.config, "auth") else None,
            )

        # Initialize session with the client
        read_stream, write_stream, *_ = await self.exit_stack.enter_async_context(
            client
        )
        self.session = await self.exit_stack.enter_async_context(
            ClientSession(read_stream, write_stream)
        )
        await self.session.initialize()


class MCPServer:
    """
    Manages connection to an MCP server and provides access to its tools.

    This class handles the connection lifecycle in a background thread, allowing
    synchronous usage of asynchronous MCP operations. Tools are automatically
    discovered and converted to Railtracks Node classes on initialization.

    The connection remains active until explicitly closed or the context manager exits.

    Usage:
        # Direct usage
        server = MCPServer(config=MCPStdioParams(command="python", args=["-m", "mcp_server_time"]))
        tools = server.tools
        server.close()

        # Context manager (recommended)
        with MCPServer(config=config) as server:
            tools = server.tools
            # Use tools...
        # Automatically closed

    Attributes:
        config: Connection configuration (MCPStdioParams or MCPHttpParams)
        tools: List of Node classes representing MCP tools (available after initialization)

    Raises:
        FileNotFoundError: If STDIO command not found
        TimeoutError: If connection setup exceeds timeout
        ConnectionError: If connection to server fails
        RuntimeError: For other setup failures
    """

    def __init__(
        self,
        config: MCPStdioParams | MCPHttpParams,
        client_session: ClientSession | None = None,
        setup_timeout: float = 30,
    ):
        """
        Initialize and connect to an MCP server.

        Connection happens in a background thread to allow sync/async bridging.
        This method blocks until connection is established or timeout occurs.

        Args:
            config: Server configuration (STDIO or HTTP)
            client_session: Optional pre-configured client session
            setup_timeout: Maximum seconds to wait for connection (default: 30)

        Raises:
            FileNotFoundError: If STDIO command not found (with helpful PATH message)
            TimeoutError: If connection setup exceeds timeout
            ConnectionError: If connection to server fails
            RuntimeError: For other setup failures
        """
        self.client = None
        self.config = config
        self.client_session = client_session
        self._tools = None
        self._loop = None
        self._setup_exception = None  # Captures exceptions from background thread
        self._thread = threading.Thread(target=self._thread_main, daemon=True)
        self._ready_event = threading.Event()
        self._shutdown_event = None
        self._thread.start()

        # Wait for background thread to complete setup
        ready = self._ready_event.wait(timeout=setup_timeout)

        if not ready:
            # Setup timed out - clean up and raise error
            self.close()
            raise TimeoutError(
                f"MCP server setup timed out after {setup_timeout} seconds. "
                f"Command: {self.config.command if isinstance(self.config, MCPStdioParams) else self.config.url}"
            )

        # If setup failed with an exception, re-raise it with enhanced context
        if self._setup_exception is not None:
            if isinstance(self._setup_exception, FileNotFoundError):
                if isinstance(self.config, MCPStdioParams):
                    raise FileNotFoundError(
                        f"MCP server command not found: '{self.config.command}'. "
                        f"Please verify the command is installed and in your PATH."
                    ) from self._setup_exception
                else:
                    raise self._setup_exception
            elif isinstance(self._setup_exception, ConnectionError):
                raise ConnectionError(
                    f"Failed to connect to MCP server: {str(self._setup_exception)}"
                ) from self._setup_exception
            else:
                raise RuntimeError(
                    f"MCP server setup failed: {type(self._setup_exception).__name__}: "
                    f"{str(self._setup_exception)}"
                ) from self._setup_exception

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()

    def _thread_main(self):
        """
        Main method for the background thread that manages the async event loop.

        Creates a new event loop, runs the setup process, and waits for shutdown signal.
        Any exceptions during setup are captured and re-raised in __init__.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self._loop = loop

        self._shutdown_event = asyncio.Event()
        try:
            self._loop.run_until_complete(self._setup())
        except asyncio.exceptions.CancelledError:
            # Ensure shutdown event is set so thread can exit cleanly
            loop.call_soon_threadsafe(self._shutdown_event.set)
            raise
        except Exception as e:
            # Capture any exception during setup for re-raising in __init__
            self._setup_exception = e
            # Set shutdown event so thread can exit
            loop.call_soon_threadsafe(self._shutdown_event.set)
        finally:
            # Signal that setup is complete (success or failure)
            self._ready_event.set()
            # Wait for shutdown signal before closing loop
            loop.run_until_complete(self._shutdown_event.wait())
            self._loop.close()

    async def _setup(self):
        """
        Initialize the MCP client connection and fetch available tools.

        This runs once in the background thread when the server is created.
        Tools are converted to Railtracks Node classes for easy integration.

        Raises:
            Exception: Any connection or initialization errors are captured
                      and re-raised in __init__ with enhanced context
        """
        self.client = MCPAsyncClient(self.config, self.client_session)
        await self.client.connect()
        tools = await self.client.list_tools()
        self._tools = [from_mcp(tool, self.client, self._loop) for tool in tools]

    def close(self):
        """
        Close the MCP server connection and clean up resources.

        Signals the background thread to shut down and waits for it to complete.
        Safe to call even if initialization failed early.
        """
        if self._loop is not None and self._shutdown_event is not None:
            self._loop.call_soon_threadsafe(self._shutdown_event.set)
            self._thread.join()

    @property
    def tools(self) -> list[Type[Node]]:
        """
        Get the list of tools available from the MCP server.

        Returns:
            List of Node classes, each representing an MCP tool.
            Each Node has:
            - name(): Tool name
            - tool_info(): Tool metadata (description, parameters)
            - prepare_tool(**kwargs): Create tool instance with arguments
        """
        return self._tools


def from_mcp(
    tool: Tool,
    client: MCPAsyncClient,
    loop: asyncio.AbstractEventLoop,
) -> Type[Node]:
    """
    Convert an MCP tool into a Railtracks Node class.

    Creates a Node subclass that bridges between Railtracks' synchronous API
    and the MCP tool's asynchronous execution. The Node can be used directly
    with rt.call() or passed to agents as a tool.

    Args:
        tool: The MCP tool object with name, description, and schema
        client: Async client for communicating with the MCP server
        loop: Event loop running in the background thread

    Returns:
        A Node subclass that:
        - Implements invoke() to call the MCP tool
        - Provides tool_info() for schema introspection
        - Supports prepare_tool() for argument binding
        - Can be used with rt.call() and agent_node()

    Example:
        server = rt.connect_mcp(config)
        ToolNode = from_mcp(tool, server.client, server._loop)
        result = await rt.call(ToolNode, param1="value1")
    """

    class MCPToolNode(Node):
        """Dynamic Node class representing an MCP tool."""

        def __init__(self, **kwargs):
            """Initialize with tool arguments."""
            super().__init__()
            self.kwargs = kwargs

        def invoke(self):
            """
            Execute the MCP tool with the provided arguments.

            Bridges from sync to async by running the tool call in the
            background thread's event loop and waiting for the result.

            Returns:
                The tool's execution result

            Raises:
                RuntimeError: If tool execution fails
            """
            try:
                future = asyncio.run_coroutine_threadsafe(
                    client.call_tool(tool.name, self.kwargs), loop
                )
                result = future.result(timeout=client.config.timeout.total_seconds())
                return result
            except Exception as e:
                raise RuntimeError(
                    f"Tool invocation failed: {type(e).__name__}: {str(e)}"
                ) from e

        @classmethod
        def name(cls):
            """Return the tool name."""
            return tool.name

        @classmethod
        def tool_info(cls) -> Tool:
            """
            Get tool metadata including parameters and description.

            Returns:
                Tool object with name, description, and parameter schemas
            """
            return Tool.from_mcp(tool)

        @classmethod
        def prepare_tool(cls, **kwargs) -> Self:
            """
            Create a tool instance with bound arguments.

            Args:
                **kwargs: Arguments to pass to the tool

            Returns:
                Tool instance ready for execution
            """
            return cls(**kwargs)

        @classmethod
        def type(cls):
            """Return the node type identifier."""
            return "Tool"

    return MCPToolNode



from typing import Literal, TypeVar

from ...providers import ModelProvider
from ._provider_wrapper import ProviderLLMWrapper

_TStream = TypeVar("_TStream", Literal[True], Literal[False])


class GeminiLLM(ProviderLLMWrapper[_TStream]):
    def full_model_name(self, model_name: str) -> str:
        # for gemini models through litellm, we need 'gemini/{model_name}' format, but we do this after the checks in ProiLLMWrapper init
        return f"gemini/{model_name}"

    @classmethod
    def model_gateway(cls):
        return ModelProvider.GEMINI  # litellm uses this for the provider for Gemini, we are using this in the checks in _provider_wrapper.py



import pytest
from railtracks.llm.tools.parameters._base import Parameter, ParameterType


def test_parameter_init_and_repr():
    p = Parameter("foo", description="desc", required=False, default="bar", enum=["bar", "baz"])
    assert p.name == "foo"
    assert p.description == "desc"
    assert not p.required
    assert p.default == "bar"
    assert p.enum == ["bar", "baz"]

def test_parameter_to_json_schema():
    p = Parameter("foo", param_type="string", description="desc", required=True, default="bar", enum=["bar", "baz"], default_present=True)
    schema = p.to_json_schema()
    assert schema["type"] == "string"
    assert schema["description"] == "desc"
    assert schema["enum"] == ["bar", "baz"]
    assert schema["default"] == "bar"

def test_param_type_from_python_type():
    assert ParameterType.from_python_type(str) == ParameterType.STRING
    assert ParameterType.from_python_type(int) == ParameterType.INTEGER
    assert ParameterType.from_python_type(float) == ParameterType.FLOAT
    assert ParameterType.from_python_type(bool) == ParameterType.BOOLEAN
    assert ParameterType.from_python_type(list) == ParameterType.ARRAY
    assert ParameterType.from_python_type(dict) == ParameterType.OBJECT
    assert ParameterType.from_python_type(type(None)) == ParameterType.NONE



"""
Test MCP Error Handling
========================

Tests that verify proper error handling during MCP server connection setup,
including invalid commands, timeouts, and exception propagation from background threads.
"""

import threading
import time

import pytest
from railtracks.rt_mcp import MCPStdioParams
from railtracks.rt_mcp.main import MCPServer


class TestConnectionErrorHandling:
    """Test proper error handling during connection setup."""

    def test_invalid_command_raises_file_not_found(self):
        """Test that invalid commands properly raise FileNotFoundError."""
        with pytest.raises(FileNotFoundError) as exc_info:
            MCPServer(
                config=MCPStdioParams(
                    command="nonexistent_command_12345",
                    args=["test"]
                ),
                setup_timeout=5
            )
        
        # Verify error message is helpful
        assert "nonexistent_command_12345" in str(exc_info.value)
        assert "PATH" in str(exc_info.value) or "not found" in str(exc_info.value).lower()

    @pytest.mark.asyncio
    async def test_setup_exception_propagation(self, stdio_config, mock_client_session, patch_stdio_client, patch_ClientSession):
        """Test that exceptions during setup are properly propagated."""
        # Create a mock that raises an exception during connect
        mock_client_session.initialize.side_effect = ConnectionError("Failed to connect")
        
        with pytest.raises(RuntimeError) as exc_info:
            server = MCPServer(
                config=stdio_config,
                client_session=None,  # Force it to create new session
                setup_timeout=5
            )
        
        # Verify the error was propagated
        assert "Failed to connect" in str(exc_info.value) or exc_info.value.__cause__ is not None


class TestErrorContextAndMessages:
    """Test that error messages provide helpful context."""

    def test_file_not_found_has_helpful_message(self):
        """Test that FileNotFoundError includes helpful troubleshooting info."""
        with pytest.raises(FileNotFoundError) as exc_info:
            MCPServer(
                config=MCPStdioParams(
                    command="missing_mcp_server",
                    args=[]
                ),
                setup_timeout=5
            )
        
        error_msg = str(exc_info.value)
        # Should mention the command name
        assert "missing_mcp_server" in error_msg
        # Should provide helpful context
        assert any(hint in error_msg for hint in ["PATH", "installed", "not found"])


class TestThreadCleanup:
    """Test that threads are properly cleaned up after errors."""

    def test_thread_cleaned_up_after_error(self):
        """Test that background thread is cleaned up after connection error."""
        initial_thread_count = threading.active_count()
        
        try:
            MCPServer(
                config=MCPStdioParams(
                    command="nonexistent_command",
                    args=[]
                ),
                setup_timeout=5
            )
        except FileNotFoundError:
            pass  # Expected
        
        # Give threads time to clean up
        time.sleep(0.5)
        
        final_thread_count = threading.active_count()
        
        # Should not leak threads (allow small variance for system threads)
        assert final_thread_count <= initial_thread_count + 2


class TestSuccessfulConnection:
    """Sanity checks that valid connections still work."""

    @pytest.mark.asyncio
    async def test_valid_connection_succeeds(self, stdio_config, mock_client_session):
        """Test that valid connections work correctly."""
        server = MCPServer(
            config=stdio_config,
            client_session=mock_client_session,
            setup_timeout=30
        )
        
        # Should have tools
        assert server.tools is not None
        
        # Clean up
        server.close()

    def test_context_manager_cleanup_after_error(self, stdio_config, mock_client_session, patch_stdio_client, patch_ClientSession):
        """Test that context manager properly cleans up after errors."""
        mock_client_session.initialize.side_effect = ConnectionError("Test error")
        
        with pytest.raises(RuntimeError):
            with MCPServer(
                config=stdio_config,
                client_session=None,  # Force new session creation
                setup_timeout=5
            ) as server:
                pass
        
        # Context manager should handle cleanup even with error


class TestCloseMethod:
    """Test that close() method handles edge cases."""

    def test_close_without_init(self):
        """Test that close() can be called even if init failed early."""
        server = object.__new__(MCPServer)
        server._loop = None
        server._shutdown_event = None
        server._thread = None
        
        # Should not raise an error
        server.close()



from __future__ import annotations

import asyncio
import random
from copy import deepcopy
from typing import List

import pytest
import railtracks as rt
from railtracks.exceptions import GlobalTimeOutError
from railtracks.nodes.nodes import Node


@pytest.mark.asyncio
async def test_message_history_not_mutated_terminal_llm(terminal_nodes):
    """
    Verify that message history is not modified after rt.call when passed to nodes constructed using different methods.
    """
    rng_node, rng_operation_node, math_detective_node = (
        terminal_nodes  # All nodes can be found in ./conftest.py
    )

    async def make_math_game_node(message_history: rt.llm.MessageHistory):
        original_message_history = deepcopy(message_history)

        # Common parameters for node calls
        call_params = {"user_input": message_history}

        # First node call
        random_num_list_response = await rt.call(rng_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 1"

        message_history.append(
            rt.llm.AssistantMessage(
                "The list of random integer: " + str(random_num_list_response)
            )
        )
        original_message_history.append(
            rt.llm.AssistantMessage(
                "The list of random integer: " + str(random_num_list_response)
            )
        )

        # Second node call
        operation_response = await rt.call(rng_operation_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 2"

        message_history.append(
            rt.llm.AssistantMessage("The result int (x) = " + str(operation_response))
        )
        original_message_history.append(
            rt.llm.AssistantMessage("The result int (x) = " + str(operation_response))
        )

        # Third node call
        response = await rt.call(math_detective_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 3"

        return response

    MathGameNode = rt.function_node(make_math_game_node)  # noqa: N806

    with rt.Session(logging_setting="NONE"):
        message_history = rt.llm.MessageHistory(
            [rt.llm.UserMessage("You can start the game")]
        )
        original_message_history = deepcopy(message_history)
        _ = await rt.call(MathGameNode, message_history=message_history)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after runner run"


@pytest.mark.asyncio
async def test_message_history_not_mutated_structured_llm(structured_nodes):
    """
    Verify that message history is not modified after rt.call when passed to nodes constructed using different methods.
    """
    math_undergrad_student_node, math_professor_node = (
        structured_nodes  # All nodes can be found in ./conftest.py
    )

    async def math_proof_node(message_history: rt.llm.MessageHistory):
        original_message_history = deepcopy(message_history)

        # Common parameters for node calls
        call_params = {"user_input": message_history}

        # First node (math student node)
        student_proof = await rt.call(math_undergrad_student_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 1"

        message_history.append(
            rt.llm.AssistantMessage("The proof: " + student_proof.structured.proof)
        )
        original_message_history.append(
            rt.llm.AssistantMessage("The proof: " + student_proof.structured.proof)
        )

        # Second node call (math professor node)
        prof_grade = await rt.call(math_professor_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 2"

        message_history.append(
            rt.llm.AssistantMessage(
                "The grade: " + str(prof_grade.structured.overall_score)
            )
        )
        message_history.append(
            rt.llm.AssistantMessage("The feedback: " + prof_grade.structured.feedback)
        )
        original_message_history.append(
            rt.llm.AssistantMessage(
                "The grade: " + str(prof_grade.structured.overall_score)
            )
        )
        original_message_history.append(
            rt.llm.AssistantMessage("The feedback: " + prof_grade.structured.feedback)
        )

        return prof_grade

    MathProofNode = rt.function_node(math_proof_node)  # noqa: N806

    with rt.Session(logging_setting="NONE"):
        message_history = rt.llm.MessageHistory(
            [
                rt.llm.UserMessage(
                    "Prove that the sum of all numbers until infinity is -1/12"
                )
            ]
        )
        original_message_history = deepcopy(message_history)
        _ = await rt.call(MathProofNode, message_history=message_history)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after runner run"


@pytest.mark.timeout(34)
@pytest.mark.asyncio
async def test_message_history_not_mutated_tool_call_llm(tool_calling_nodes):
    """
    Verify that message history is not modified after rt.call when passed to nodes constructed using different methods.
    """
    currrency_converter_node, travel_planner_node = (
        tool_calling_nodes  # All nodes can be found in ./conftest.py
    )

    async def travel_summarizer_node(message_history: rt.llm.MessageHistory):
        original_message_history = deepcopy(message_history)

        # Common parameters for node calls
        call_params = {"user_input": message_history}

        # First node call
        travel_planner_response = await rt.call(travel_planner_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 1"

        message_history.append(
            rt.llm.AssistantMessage("The travel plan: " + str(travel_planner_response))
        )
        original_message_history.append(
            rt.llm.AssistantMessage("The travel plan: " + str(travel_planner_response))
        )

        # Second node call
        response = await rt.call(currrency_converter_node, **call_params)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after rt.call 2"

        return response

    TravelSummarizerNode = rt.function_node(travel_summarizer_node)  # noqa: N806
    with rt.Session(logging_setting="NONE"):
        message_history = rt.llm.MessageHistory(
            [
                rt.llm.UserMessage(
                    "I want to plan a trip to from Delhi to New York for a week. Please provide me with a budget summary for the trip."
                )
            ]
        )
        original_message_history = deepcopy(message_history)
        _ = await rt.call(TravelSummarizerNode, message_history=message_history)
        assert all(
            orig.content == new.content
            for orig, new in zip(original_message_history, message_history)
        ), "Message history modified after runner run"


async def test_no_context_call():
    with pytest.raises(Exception):
        await rt.call(
            lambda: "This should not work",
            "This is a test argument",
            key="This is a test keyword argument",
        )

def add(x: float, y: float):
    """A simple synchronous function that adds two numbers."""
    return x + y


AddNode = rt.function_node(add)


async def async_add_many(pairs: list[float]):
    """An asynchronous function that adds many numbers."""
    total = 0
    for i in range(len(pairs)):
        total = await rt.call(AddNode, total, pairs[i])
    return total


AddManyAsyncNode = rt.function_node(async_add_many)


# ============================================ START Many calls and Timeout tests ============================================

RNGNode = rt.function_node(random.random)


async def many_calls(num_calls: int, parallel_calls: int):
    data = []
    for _ in range(num_calls):
        contracts = [rt.call(RNGNode) for _ in range(parallel_calls)]
        results = await asyncio.gather(*contracts)
        data.extend(results)
    return data


ManyCalls = rt.function_node(many_calls)


@pytest.mark.asyncio
async def many_calls_tester(num_calls: int, parallel_calls: int):
    with rt.Session(logging_setting="NONE") as run:
        finished_result = await rt.call(ManyCalls, num_calls, parallel_calls)
        info = run.info

    ans = finished_result

    assert isinstance(ans, list)
    assert len(ans) == num_calls * parallel_calls
    assert all(0 < x < 1 for x in ans)
    assert {x.step for x in info.all_stamps} == set(
        range(num_calls * parallel_calls * 2 + 2)
    )

    assert len(info.all_stamps) == 2 * num_calls * parallel_calls + 2


@pytest.mark.timeout(5)
@pytest.mark.asyncio
async def test_no_deadlock():
    num_calls = 4
    parallel_calls = 55

    await many_calls_tester(num_calls, parallel_calls)


@pytest.mark.asyncio
async def test_small_no_deadlock():
    num_calls = 10
    parallel_calls = 15

    await many_calls_tester(num_calls, parallel_calls)


@pytest.mark.asyncio
async def test_large_no_deadlock():
    num_calls = 45
    parallel_calls = 23

    await many_calls_tester(num_calls, parallel_calls)


@pytest.mark.asyncio
async def test_simple_rng():
    with rt.Session(logging_setting="NONE"):
        result = await rt.call(RNGNode)

    assert 0 < result < 1


class NestedManyCalls(Node):
    def __init__(self, num_calls: int, parallel_calls: int, depth: int):
        self.num_calls = num_calls
        self.parallel_calls = parallel_calls
        self.depth = depth
        super().__init__()

    async def invoke(
        self,
    ):
        data = []
        for _ in range(self.num_calls):
            if self.depth == 0:
                contracts = [rt.call(RNGNode) for _ in range(self.parallel_calls)]
                results = await asyncio.gather(*contracts)

            else:
                contracts = [
                    rt.call(
                        NestedManyCalls,
                        self.num_calls,
                        self.parallel_calls,
                        self.depth - 1,
                    )
                    for _ in range(self.parallel_calls)
                ]

                results = await asyncio.gather(*contracts)
                # flatten the list here.
                results = [x for y in results for x in y]
            data.extend(results)
        return data

    @classmethod
    def name(cls) -> str:
        return "NestedManyCalls"

    @classmethod
    def type(cls):
        return "Tool"


@pytest.mark.asyncio
async def nested_many_calls_tester(num_calls: int, parallel_calls: int, depth: int):
    with rt.Session(logging_setting="NONE") as run:
        await rt.call(NestedManyCalls, num_calls, parallel_calls, depth)

    ans = run.info

    assert isinstance(ans.answer, list)
    assert len(ans.answer) == (parallel_calls * num_calls) ** (depth + 1)
    assert all(0 < x < 1 for x in ans.answer)

    r_h = ans.request_forest
    assert len(r_h.insertion_request) == 1
    child_requests = r_h.children(r_h.insertion_request[0].sink_id)

    assert len(child_requests) == num_calls * parallel_calls
    for r in child_requests:
        assert r.input[0][0] == num_calls
        assert r.input[0][1] == parallel_calls
        assert 0 < r.input[0][2] < depth


@pytest.mark.timeout(4)
@pytest.mark.asyncio
async def test_nested_no_deadlock():
    num_calls = 2
    parallel_calls = 2
    depth = 3

    await nested_many_calls_tester(num_calls, parallel_calls, depth)


@pytest.mark.asyncio
async def test_nested_no_deadlock_harder():
    num_calls = 1
    parallel_calls = 3
    depth = 3

    await nested_many_calls_tester(num_calls, parallel_calls, depth)


@pytest.mark.asyncio
async def test_nested_no_deadlock_harder_2():
    num_calls = 3
    parallel_calls = 1
    depth = 3

    await nested_many_calls_tester(num_calls, parallel_calls, depth)


@pytest.mark.asyncio
async def test_multiple_runs():
    with rt.Session(logging_setting="NONE") as run:
        result = await rt.call(RNGNode)
        assert 0 < result < 1

        result = await rt.call(RNGNode)

        info = run.info
        assert isinstance(info.answer, List)
        assert 0 < info.answer[0] < 1
        assert 0 < info.answer[1] < 1

        insertion_requests = info.request_forest.insertion_request

        assert isinstance(insertion_requests, List)
        assert len(insertion_requests) == 2
        for i_r in insertion_requests:
            i_r_id = i_r.identifier

            subset_info = info._get_info(i_r_id)
            assert 0 < subset_info.answer < 1
            assert len(subset_info.node_forest.heap()) == 1


@pytest.mark.asyncio
async def test_multiple_runs_async():
    with rt.Session(logging_setting="NONE") as run:
        result = await rt.call(RNGNode)
        assert 0 < result < 1

        result = await rt.call(RNGNode)

        result = run.info.answer
        assert isinstance(result, List)
        assert 0 < result[0] < 1
        assert 0 < result[1] < 1

        info = run.info

        insertion_requests = info.request_forest.insertion_request

        assert isinstance(insertion_requests, List)
        assert len(insertion_requests) == 2
        for i_r in insertion_requests:
            i_r_id = i_r.identifier

            subset_info = info._get_info(i_r_id)
            assert 0 < subset_info.answer < 1
            assert len(subset_info.node_forest.heap()) == 1


def level_3(message: str):
    return message


Level3 = rt.function_node(level_3)


async def a_level_2(message: str):
    return await rt.call(Level3, message)


ALevel2 = rt.function_node(a_level_2)


@pytest.mark.parametrize("level_2_node", [ALevel2], ids=["async"])
@pytest.mark.asyncio
async def test_multi_level_calls(level_2_node):
    async def level_1_async(message: str):
        return await rt.call(level_2_node, message)

    ALevel1 = rt.function_node(level_1_async)  # noqa: N806

    with rt.Session(logging_setting="NONE"):
        result = await rt.call(ALevel1, "Hello from Level 1 (async)")
        assert result == "Hello from Level 1 (async)"

    with rt.Session(logging_setting="NONE"):
        result = await rt.call(ALevel1, "Hello from Level 1 (async)")
        assert result == "Hello from Level 1 (async)"


async def timeout_node(timeout_len: float):
    """
    A node that sleeps for the given timeout length.
    """
    await asyncio.sleep(timeout_len)
    return timeout_len


TimeoutNode = rt.function_node(timeout_node)


@pytest.mark.asyncio
async def test_timeout():
    with rt.Session(logging_setting="NONE", timeout=0.1):
        with pytest.raises(GlobalTimeOutError):
            await rt.call(TimeoutNode, 0.3)


async def timeout_thrower():
    raise asyncio.TimeoutError("Test timeout error")


TimeoutThrower = rt.function_node(timeout_thrower)


@pytest.mark.asyncio
async def test_timeout_thrower():
    with rt.Session(logging_setting="NONE"):
        try:
            await rt.call(TimeoutThrower)
        except Exception as e:
            assert isinstance(e, asyncio.TimeoutError)


# ============================================ END Many calls and Timeout tests ============================================



"""
Docstring parsing utilities.

This module contains functions for parsing Python docstrings to extract
parameter descriptions and other documentation.
"""

import re
from typing import Dict

from .parameters import Parameter, ParameterType


# HELPER
def param_from_python_type(
    py_type, name: str = "", description: str | None = None, required: bool = True
) -> Parameter:
    mapped_type = ParameterType.from_python_type(py_type).value
    return Parameter(
        name=name, param_type=mapped_type, description=description, required=required
    )


def parse_docstring_args(docstring: str) -> Dict[str, str]:
    """
    Parses the 'Args:' section from a docstring.
    Returns a dictionary mapping parameter names to their descriptions.

    Args:
        docstring: The docstring to parse.

    Returns:
        A dictionary mapping parameter names to their descriptions.
    """
    if not docstring:
        return {}

    # Extract the Args section
    args_section = extract_args_section(docstring)
    if not args_section:
        return {}

    # Parse the arguments and their descriptions
    return parse_args_section(args_section)


def extract_args_section(docstring: str) -> str:
    """
    Extracts the 'Args:' section from a docstring.

    Args:
        docstring: The docstring to extract from.

    Returns:
        The extracted 'Args:' section as a string, or an empty string if not found.
    """
    args_section = ""
    split_lines = docstring.splitlines()

    # Find the Args: section
    in_args_section = False
    for i, line in enumerate(split_lines):
        if line.strip().startswith("Args:"):
            in_args_section = True
            # Skip the "Args:" line itself
            continue

        if in_args_section:
            # Check if we've reached another section (e.g., "Returns:")
            if (
                line.strip()
                and line.strip().endswith(":")
                and not line.strip().startswith(" ")
            ):
                break

            # Add the line to our args section
            args_section += line + "\n"

    return args_section


def parse_args_section(args_section: str) -> Dict[str, str]:
    """
    Parses an 'Args:' section into a dictionary of parameter names and descriptions.

    Args:
        args_section: The extracted 'Args:' section text.

    Returns:
        A dictionary mapping parameter names to their descriptions.
    """
    # Regular expression to match parameter definitions
    # This handles both formats:
    # - param_name: Description
    # - param_name (type): Description
    pattern = re.compile(r"^\s*(\w+)(?:\s*\([^)]+\))?:\s*(.+)$")

    arg_descriptions = {}
    current_arg = None
    current_description = []

    for line in args_section.splitlines():
        # Skip empty lines
        if not line.strip():
            continue

        # Check if this is a new parameter definition
        match = pattern.match(line)
        if match:
            # If we were processing a previous parameter, save it
            if current_arg and current_description:
                arg_descriptions[current_arg] = " ".join(current_description).strip()

            # Start a new parameter
            arg_name, arg_desc = match.groups()
            current_arg = arg_name
            current_description = [arg_desc.strip()]
        elif current_arg:
            # This is a continuation of the previous parameter's description
            current_description.append(line.strip())

    if current_arg and current_description:
        arg_descriptions[current_arg] = " ".join(current_description).strip()

    return arg_descriptions


def extract_main_description(docstring: str) -> str:
    """
    Extracts the main description from a docstring (before any sections like Args:, Returns:, etc.)

    Args:
        docstring: The docstring to extract from.

    Returns:
        The main description as a string.
    """
    if not docstring:
        return ""

    # Split the docstring into lines
    lines = docstring.splitlines()

    # Collect lines until we hit a section marker (like "Args:")
    main_description = []
    for line in lines:
        if (
            line.strip()
            and line.strip().endswith(":")
            and not line.strip().startswith(" ")
        ):
            break
        main_description.append(line)

    return "\n".join(main_description).strip()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Callable, Dict, List, ParamSpec, Tuple, TypeVar

from ..context.central import update_parent_id
from ..execution.coordinator import Coordinator
from ..execution.task import Task
from ..pubsub.messages import (
    FatalFailure,
    RequestCompletionMessage,
    RequestCreation,
    RequestCreationFailure,
    RequestFailure,
    RequestFinishedBase,
    RequestSuccess,
)
from ..pubsub.publisher import RTPublisher
from ..utils.logging.action import (
    RequestCreationAction,
    RequestFailureAction,
    RequestSuccessAction,
)

# all the things we need to import from RT directly.
# all the things we need to import from RT directly.
from .request import Cancelled, Failure
from .utils import create_sub_state_info

if TYPE_CHECKING:
    from railtracks.utils.config import ExecutorConfig

from railtracks.exceptions import FatalError, NodeInvocationError
from railtracks.nodes.nodes import Node
from railtracks.utils.logging.create import get_rt_logger
from railtracks.utils.profiling import Stamp

from .info import ExecutionInfo

_TOutput = TypeVar("_TOutput")
_P = ParamSpec("_P")


class RTState:
    """
    RTState is an internal RT object used to manage state of the request completion system. This object has a couple of
    major functions:
    1. It allows you to create a new state object every time you want to run the system.
    2. The state object will connect to the provided publisher and will handle all the messages that are sent to it.
    2. It allows you to invoke the graph at a given node and will handle the execution of the graph.
    3. It will handle all the exceptions that are thrown during the execution of the graph.
    4. It will handle the creation of new requests and the management of the state of the nodes.
    5. It will save all the details in the object so that you can access them later.
    6. Simple logging of action completed in the system.

    The object is designed to be a large state object that contains all important details about the running of the
    system during execution. It can be accessed after run, to collect important details that will aid debugging and
    allow for retrospective analysis of the system.
    """

    def __init__(
        self,
        execution_info: ExecutionInfo,
        executor_config: ExecutorConfig,
        coordinator: Coordinator,
        publisher: RTPublisher,
    ):
        self._node_heap = execution_info.node_forest
        self._request_heap = execution_info.request_forest
        self._stamper = execution_info.stamper

        self.executor_config = executor_config
        self.rc_coordinator = coordinator

        # each new instance of a state object should have its own logger.
        self.logger = get_rt_logger()

        publisher.subscribe(self.handle, "State Object Handler")
        self.publisher = publisher

    async def handle(self, item: RequestCompletionMessage) -> None:
        """
        A simpler handler function designed to handle the messages from the publisher object by the state object.

        Args:
            item: The item that was sent to the publisher. It can be any of the RequestCompletionMessage types.
        """
        # TODO: fix up these abstractions so it consistent that we are mapping the request type into its proper type.
        if isinstance(item, RequestFinishedBase):
            await self.handle_result(item)
        if isinstance(item, RequestCreation):
            if item.current_node_id is not None:
                update_parent_id(item.current_node_id, item.current_run_id)

            assert item.new_request_id not in self._request_heap.heap().keys()

            await self.call_nodes(
                parent_node_id=item.current_node_id,
                request_id=item.new_request_id,
                node=item.new_node_type,
                args=item.args,
                kwargs=item.kwargs,
            )

    def shutdown(self):
        """
        Shutdown the state object and all of its references.
        """
        self.rc_coordinator.shutdown()

    @property
    def is_empty(self):
        """
        Determines if the state object is currently empty (i.e. no nodes or requests are present).
        """
        return len(self._node_heap.heap()) == 0 and len(self._request_heap.heap()) == 0

    def add_stamp(self, message: str):
        """
        Manually adds a stamp with the provided message to the stamp generator

        Args:
            message: The message you would like to attach to the stamp

        """
        self._stamper.create_stamp(message)

    async def cancel(self, node_id: str):
        """
        Cancels the running process of the node with the given identifier.

        It will update the state, but it will not kill the process running the node.
        """
        if node_id not in self._node_heap.heap():
            assert False

        r_id = self._request_heap.get_request_from_child_id(node_id)
        self._request_heap.update(
            r_id, Cancelled, self._stamper.create_stamp(f"Cancelled request {r_id}")
        )

    def _create_node_and_request(
        self,
        *,
        parent_node_id: str,
        request_id: str | None,
        node: Callable[_P, Node],
        args: _P.args,
        kwargs: _P.kwargs,
    ) -> str:
        """
        Creates a node using the creator function (node).

        ASIDE: This function is the only way you should create nodes in the system. If you create them elsehwere they
        will not be tracked in the system.


        Args:
            parent_node_id: The parent node id of the node you are creating (used for tracking of requests)
            node: The Node you would like to create
            args: The arguments to pass to the node
            kwargs: The keyword arguments to pass to the node

        Returns:
            The request id of the request created between parent and child.
        """

        # 1. Create the node here
        node = node(*args, **kwargs)

        # 2. Add it to the node heap.
        sc = self._stamper.stamp_creator()
        parent_node_type = self._node_heap.get_node_type(parent_node_id)
        parent_node_name = parent_node_type.name() if parent_node_type else "START"

        request_creation_obj = RequestCreationAction(
            parent_node_name=parent_node_name,
            child_node_name=node.name(),
            input_args=args,
            input_kwargs=kwargs,
        )

        stamp = sc(request_creation_obj.to_logging_msg())

        self._node_heap.update(node, stamp)

        # 3. Attach the requests that will tied to this node.
        request_ids = self._create_new_request_set(
            parent_node=parent_node_id,
            children=[node.uuid],
            input_args=[args],
            input_kwargs=[kwargs],
            stamp=stamp,
            request_ids=[request_id],
        )

        self.logger.info(request_creation_obj.to_logging_msg())
        # 4. Return the request id of the node that was created.
        return request_ids[0]

    async def call_nodes(
        self,
        *,
        parent_node_id: str | None,
        request_id: str | None,
        node: Callable[_P, Node[_TOutput]],
        args: _P.args,
        kwargs: _P.kwargs,
    ):
        """
        This function will handle the creation of the node and the subsequent running of the node returning the result.

        Important: If an exception was thrown during the execution of the request, it will pass through this function,
        and will be raised.

        Args:
            parent_node_id: The parent node id of the node you are calling. None if it has no parent.
            request_id: The request id of the request you are creating. If None, a new request id will be created.
            node: The node you would like to create.
            args: The arguments to pass to the node.
            kwargs: The keyword arguments to pass to the node.

        Returns:
            The output of the node that was run. It will match the output type of the child node that was run.

        """
        try:
            request_id = self._create_node_and_request(
                parent_node_id=parent_node_id,
                request_id=request_id,
                node=node,
                args=args,
                kwargs=kwargs,
            )
        except Exception as e:
            # TODO improve this so we know the name of the node trying to be created in the case of a tool call llm.
            rfa = RequestFailureAction(
                node_name=node.name() if hasattr(node, "name") else "Unknown",
                exception=e,
            )
            await self.publisher.publish(
                RequestCreationFailure(
                    request_id=request_id,
                    error=e,
                )
            )
            self.logger.exception(rfa.to_logging_msg())
            raise e
        # you have to run this in a task so it isn't blocking other completions
        outputs = asyncio.create_task(self._run_request(request_id))

        return outputs

    # TODO handle the business around parent node with automatic checkpointing.
    def _create_new_request_set(
        self,
        parent_node: str | None,
        children: List[str],
        input_args: List[Tuple],
        input_kwargs: List[Dict[str, Tuple]],
        stamp: Stamp,
        request_ids: List[str | None] | None = None,
    ) -> List[str]:
        """
        Creates a new set of requests from the provided details. Essentially we will be creating a new request from the
        parent node to each of the children nodes. They will all share the stamp and will have a descriptive message
        attached to each of them.

        Note that you are supposed to provide a list of items. It is assumed that all the lists are the same length.
        Each index in the lists corresponds to the same child node.

        Args:
            parent_node: The identifier of the parent node which is calling the children. If none is provided we assume there is no parent.
            children: The list of node_ids that you would like to call.
            input_args: The list of arguments to pass to each of the children nodes.
            input_kwargs: The list of keyword arguments to pass to each of the children nodes.
            request_ids: The list of request ids to use for each of the children nodes. I
            stamp: The stamp you would like to attach to each of the requests.
        """

        # note it is assumed that all of the children id are valid and have already been created.
        assert all(n in self._node_heap for n in children), (
            "You cannot add a request for a node which has not yet been added"
        )

        if request_ids is None:
            request_ids = [None] * len(children)

        # to simplify we are going to create a new request for each child node with the parent as its source.
        request_ids = list(
            map(
                self._request_heap.create,
                request_ids,
                [parent_node] * len(children),
                children,
                input_args,
                input_kwargs,
                [stamp for _ in children],
            )
        )

        return request_ids

    async def _run_request(self, request_id: str):
        """
        Runs the request for the given request id.

        1. It will use the request to collect the identifier of the child node and then run the node.
        2. It will submit the task to the coordinator to run the node.
        3. It will return once the request has been placed.


        Args:
            request_id: The identifier for the request you would like to run


        """
        child_node_id = self._request_heap[request_id].sink_id
        node = self._node_heap[child_node_id].node
        return await self.rc_coordinator.submit(
            task=Task(request_id=request_id, node=node),
            mode="async",
        )

    async def _handle_failed_request(
        self, failed_node_name: str, request_id: str, exception: Exception
    ):
        """
        Handles the provided exception for the given request identifier. It will publish a message to the publisher if
        deemed necessary and will return a Failure object containing the exception.

        This function will handle the following tasks:
        1. logs the exception
        2. Determines if the error is fatal and whether it should kill the system. If so it will publish a FatalFailure
            message to the publisher
        3. Returns a Failure object.

        Args:
            failed_node_name: The name of the node that failed. This is used for logging and debugging.
            request_id: The request in which this error occurred
            exception: The exception thrown during this request.

        Returns:
            An object containing the error thrown during the request wrapped in a Failure object.
        """
        # before doing any handling we must make sure our exception history object is up to date.

        node_exception_action = RequestFailureAction(
            node_name=failed_node_name,
            exception=exception,
        )

        if self.executor_config.end_on_error:
            self.logger.critical(
                node_exception_action.to_logging_msg(), exc_info=exception
            )
            await self.publisher.publish(FatalFailure(error=exception))
            return Failure(exception)

        # fatal exceptions should only be thrown if there is something seriously wrong. At the moment only NodeInvocatioErrors have 'fatal' flags
        if (
            isinstance(exception, NodeInvocationError) and exception.fatal
        ) or isinstance(exception, FatalError):
            self.logger.critical(
                node_exception_action.to_logging_msg(), exc_info=exception
            )
            await self.publisher.publish(FatalFailure(error=exception))
            return Failure(exception)

        # for any other error we want it to bubble up so the user can handle.
        self.logger.error(node_exception_action.to_logging_msg(), exc_info=exception)
        return Failure(exception)

    @property
    def info(self):
        """Returns the current state as an ExecutionInfo object."""
        return ExecutionInfo(
            self._request_heap,
            self._node_heap,
            self._stamper,
        )

    def get_info(self, ids: List[str] | str) -> ExecutionInfo:
        """
        Gets the info object for the current state but filtered to include on the children of the provided ids.

        You should provide minimal ids to this function (i.e. you should not provide the list of children instead just
        the parent node id). If you provide a parent and its children behavior is undefined.
        """
        filtered_nodes, filtered_requests = create_sub_state_info(
            self._node_heap.heap(), self._request_heap.heap(), ids
        )
        # TODO: deal with the weirdness around double representation in the stamper.
        #  specifically we need to make sure that the data in the stamper is only for the subset and not for the global state.
        return ExecutionInfo(
            node_forest=filtered_nodes,
            request_forest=filtered_requests,
            stamper=self._stamper,
        )

    async def _handle_successful_request(
        self,
        node_name: str,
        result: Any,
    ):
        """
        Handles the successful completion of a request. It will log the success and return the result.
        """
        request_completion_obj = RequestSuccessAction(
            node_name=node_name,
            output=result,
        )

        self.logger.info(request_completion_obj.to_logging_msg())
        return result

    async def handle_result(self, result: RequestFinishedBase):
        """
        Handles the given request completion.

        If it is a Failure it will log and post any relevant items to the publisher and return the error.

        This function will also handle successful requests in the same way.

        Raises:
            TypeError: If the provided result does not match any of the expected types.
        """
        if isinstance(result, RequestFailure):
            # if the node state is None, it means the node was never created so we don't need to handle it
            output = await self._handle_failed_request(
                result.node.name(), result.request_id, result.error
            )
            returnable_result = result.error

        elif isinstance(result, RequestSuccess):
            output = await self._handle_successful_request(
                node_name=result.node.name(),
                result=result.result,
            )
            returnable_result = result.result

        elif isinstance(result, RequestCreationFailure):
            # we do not need to both handling this.
            return
        else:
            raise TypeError(f"Unknown result type: {type(result)}")

        stamp = self._stamper.create_stamp(f"Finished executing {result.node.name()}")

        self._request_heap.update(result.request_id, output, stamp)
        self._node_heap.update(result.node, stamp)

        return returnable_result



import os
from typing import Optional

from charset_normalizer import from_path

_pdfplumber = None


class MediaParser:
    """General-purpose media parser capable of extracting text from various file types.

    Currently supports:
        - .txt
        - .pdf
    """

    @classmethod
    def get_text(cls, path: str, **kwargs) -> str:
        """Return cleaned text extracted from a supported file.

        Args:
            path: Path to the file
            **kwargs: Parser-specific arguments (e.g., encoding for .txt files)
        """
        ext = cls._get_extension(path)
        if ext == ".txt":
            handler_name = "_parse_txt"

        elif ext == ".pdf":
            handler_name = "_parse_pdf"

        else:
            raise ValueError(f"Unsupported file type: {ext}")

        parser_function = getattr(cls, handler_name)
        raw_text = parser_function(path, **kwargs)  # Pass kwargs through
        return cls._clean_text(raw_text)

    @staticmethod
    def _parse_txt(filepath: str, encoding: Optional[str] = None, **kwargs) -> str:
        """Extract text from a plain .txt file."""
        if not os.path.isfile(filepath):
            raise FileNotFoundError(f"File not found: {filepath}")

        if encoding is not None:
            with open(filepath, "r", encoding=encoding) as f:
                return f.read()

        # Auto-detect encoding
        detected = from_path(filepath).best()
        if detected is None:
            raise ValueError(f"Failed to detect encoding for: {filepath}")

        with open(filepath, "r", encoding=detected.encoding) as f:
            return f.read()

    @staticmethod
    def _parse_pdf(filepath: str, **kwargs) -> str:
        """Extract text from a PDF using pdfplumber."""
        if not os.path.isfile(filepath):
            raise FileNotFoundError(f"File not found: {filepath}")

        global _pdfplumber

        if _pdfplumber is None:
            try:
                import pdfplumber

                _pdfplumber = pdfplumber
            except ImportError:
                raise RuntimeError(
                    "pdfplumber is required for PDF parsing but isn't installed. "
                    "Install it via `pip install pdfplumber`."
                )

        with _pdfplumber.open(filepath) as doc:
            extracted = []
            for page in doc.pages:
                text = page.extract_text()
                if text:
                    extracted.append(text)
            return "\n".join(extracted)

    @staticmethod
    def _clean_text(text: str) -> str:
        """Remove null bytes / non-printable characters while preserving whitespace."""
        if not text:
            return ""
        return "".join(char for char in text if char.isprintable() or char in "\t\n\r")

    @staticmethod
    def _get_extension(path: str) -> str:
        """Return file extension in lowercase."""
        _, ext = os.path.splitext(path)
        return ext.lower()



import asyncio

import pytest

import railtracks as rt

import sys

from railtracks.rt_mcp.main import MCPHttpParams, MCPStdioParams


@pytest.mark.skip(reason="Skipped due to LLM stochasticity")
def test_from_mcp_server_with_llm():
    time_server = rt.connect_mcp(
        MCPStdioParams(
            command=sys.executable,
            args=["-m", "mcp_server_time", "--local-timezone=America/Vancouver"],
        )
    )
    parent_tool = rt.agent_node(
        tool_nodes={*time_server.tools},
        name="Parent Tool",
        system_message=(
            "Provide a response using the tool when asked. If the tool doesn't work,"
            " respond with 'It didn't work!'"
        ),
        llm=rt.llm.OpenAILLM("gpt-4o"),
    )

    # Run the parent tool
    with rt.Session(
        logging_setting="NONE", timeout=1000
    ):
        message_history = rt.llm.MessageHistory(
            [rt.llm.UserMessage("What time is it?")]
        )
        response = asyncio.run(rt.call(parent_tool, user_input=message_history))

    assert response is not None
    assert response.content != "It didn't work!"


@pytest.mark.skip(reason="Skipped due to LLM stochasticity")
def test_from_mcp_server_with_http():
    time_server = rt.connect_mcp(MCPHttpParams(url="https://mcp.deepwiki.com/sse"))
    parent_tool = rt.agent_node(
        tool_nodes={*time_server.tools},
        name="Parent Tool",
        system_message=(
            "Provide a response using the tool when asked. If the tool doesn't work,"
            " respond with 'It didn't work!'"
        ),
        llm=rt.llm.OpenAILLM("gpt-4o"),
    )

    # Run the parent tool
    with rt.Session(
        logging_setting="NONE", timeout=1000
    ):
        message_history = rt.llm.MessageHistory(
            [rt.llm.UserMessage("Tell me about the website conductr.ai")]
        )
        response = asyncio.run(rt.call(parent_tool, user_input=message_history))

    assert response is not None
    assert response.content is not "It didn't work!"


from .errors import (
    ContextError,
    FatalError,
    GlobalTimeOutError,
    LLMError,
    NodeCreationError,
    NodeInvocationError,
)

__all__ = [
    "FatalError",
    "NodeCreationError",
    "NodeInvocationError",
    "GlobalTimeOutError",
    "LLMError",
    "ContextError",
]



import asyncio
from typing import Any, Callable, Coroutine, Union

from .messages import RequestCompletionMessage, Streaming


def stream_subscriber(
    sub_callback: Callable[[Any], Union[None, Coroutine[None, None, None]]],
) -> Callable[[RequestCompletionMessage], Coroutine[None, None, None]]:
    """
    Converts the basic streamer callback into a broadcast_callback handler designed to take in `RequestCompletionMessage`
    """

    async def subscriber_handler(item: RequestCompletionMessage):
        if isinstance(item, Streaming):
            result = sub_callback(item.streamed_object)
            if asyncio.iscoroutine(result):
                await result

    return subscriber_handler



from __future__ import annotations

import asyncio
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    ParamSpec,
    TypeVar,
)

if TYPE_CHECKING:
    from railtracks.built_nodes.easy_usage_wrappers import (
        _AsyncNodeAttachedFunc,
        _SyncNodeAttachedFunc,
    )
from railtracks.nodes.nodes import Node

from ._call import call

_P = ParamSpec("_P")
_TOutput = TypeVar("_TOutput")


async def call_batch(
    node: Callable[..., Node[_TOutput]]
    | Callable[..., _TOutput]
    | _AsyncNodeAttachedFunc[_P, _TOutput]
    | _SyncNodeAttachedFunc[_P, _TOutput],
    *iterables: Iterable[Any],
    return_exceptions: bool = True,
):
    """
    Complete a node over multiple iterables, allowing for parallel execution.

    Note the results will be returned in the order of the iterables, not the order of completion.

    If one of the nodes returns an exception, the thrown exception will be included as a response.

    Args:
        node: The node type to create.
        *iterables: The iterables to map the node over.
        return_exceptions: If True, exceptions will be returned as part of the results.
            If False, exceptions will be raised immediately, and you will lose access to the results.
            Defaults to true.

    Returns:
        An iterable of results from the node.

    Usage:
        ```python
        results = await batch(NodeA, ["hello world"] * 10)
        for result in results:
            handle(result)
        ```
    """
    # this is big typing disaster but there is no way around it. Try if if you want to.
    contracts = [call(node, *args) for args in zip(*iterables)]

    results = await asyncio.gather(*contracts, return_exceptions=return_exceptions)
    return results



from typing import Generator
import pytest
from railtracks.llm.models._litellm_wrapper import (
    _parameters_to_json_schema,
    _to_litellm_tool,
)
from railtracks.exceptions import NodeInvocationError, LLMError
from railtracks.llm import AssistantMessage, UserMessage
from pydantic import BaseModel
from railtracks.llm.response import Response
from json import JSONDecodeError
import litellm
from railtracks.llm.content import Stream
import json

class TestHelpers:

    # =================================== START _parameters_to_json_schema Tests ==================================
    # parameters_to_json_schema is guaranteed to get only a set of Parameter objects
    def test_parameters_to_json_schema_with_parameters_set(
        self, tool_with_parameters_set
    ):
        """
        Test _parameters_to_json_schema with a set of Parameter objects.
        """
        schema = _parameters_to_json_schema(tool_with_parameters_set.parameters)
        assert schema["type"] == "object"
        assert "properties" in schema
        assert "param1" in schema["properties"]
        assert schema["properties"]["param1"]["type"] == "string"
        assert schema["properties"]["param1"]["description"] == "A string parameter."
        assert "required" in schema
        assert "param1" in schema["required"]

    def test_parameters_to_json_schema_with_empty_set(self):
        schema = _parameters_to_json_schema(set())
        assert schema == {"type": "object", "properties": {}}

    def test_parameters_to_json_schema_invalid_input(self):
        """
        Test _parameters_to_json_schema with invalid input.
        """
        with pytest.raises(NodeInvocationError):
            _parameters_to_json_schema(123)  # type: ignore

    # =================================== END _parameters_to_json_schema Tests ====================================

    # =================================== START _to_litellm_tool Tests ==================================
    def test_to_litellm_tool(self, tool):
        """
        Test _to_litellm_tool with a valid Tool instance.
        """
        litellm_tool = _to_litellm_tool(tool)
        assert litellm_tool["type"] == "function"
        assert "function" in litellm_tool
        assert litellm_tool["function"]["name"] == "example_tool"
        assert litellm_tool["function"]["description"] == "This is an example tool."
        assert "parameters" in litellm_tool["function"]

    # =================================== END _to_litellm_tool Tests ====================================

    # =================================== START _to_litellm_message Tests ==================================
    def test_to_litellm_message_user_message(self, mock_litellm_wrapper, user_message):
        """
        Test _to_litellm_message with a UserMessage instance.
        """
        wrapper = mock_litellm_wrapper()
        litellm_message = wrapper._to_litellm_message(user_message)
        assert litellm_message["role"] == "user"
        assert litellm_message["content"] == "This is a user message."

    def test_to_litellm_message_assistant_message(self, mock_litellm_wrapper, assistant_message):
        """
        Test _to_litellm_message with an AssistantMessage instance.
        """
        wrapper = mock_litellm_wrapper()
        litellm_message = wrapper._to_litellm_message(assistant_message)
        assert litellm_message["role"] == "assistant"
        assert litellm_message["content"] == "This is an assistant message."

    def test_to_litellm_message_tool_message(self, mock_litellm_wrapper, tool_message):
        """
        Test _to_litellm_message with a ToolMessage instance.
        """
        wrapper = mock_litellm_wrapper()
        litellm_message = wrapper._to_litellm_message(tool_message)
        assert litellm_message["role"] == "tool"
        assert litellm_message["name"] == "example_tool"
        assert litellm_message["tool_call_id"] == "123"
        assert litellm_message["content"] == "success"

    def test_to_litellm_message_tool_call_list(self, mock_litellm_wrapper, tool_call):
        """
        Test _to_litellm_message with a list of ToolCall instances.
        """
        tool_calls = [tool_call]
        message = AssistantMessage(content=tool_calls)
        wrapper = mock_litellm_wrapper()
        litellm_message = wrapper._to_litellm_message(message)
        assert litellm_message["role"] == "assistant"
        assert len(litellm_message["tool_calls"]) == 1
        assert litellm_message["tool_calls"][0].function.name == "example_tool"

    def test_to_litellm_message_user_message_with_attachments(
        self,
        mock_litellm_wrapper,
    ):
        """
        Test _to_litellm_message handles multimodal user messages with attachments.
        """
        wrapper = mock_litellm_wrapper()
        attachment_data_uri = "data:image/png;base64,iVBORw0KGgo="
        message = UserMessage(
            content="View this image.",
            attachment=[attachment_data_uri],
        )

        litellm_message = wrapper._to_litellm_message(message)

        assert litellm_message["role"] == "user"
        assert isinstance(litellm_message["content"], list)
        assert litellm_message["content"][0] == {
            "type": "text",
            "text": "View this image.",
        }
        assert litellm_message["content"][1] == {
            "type": "image_url",
            "image_url": {"url": attachment_data_uri},
        }

    # =================================== END _to_litellm_message Tests ====================================


# ================= BEGIN str/model_name (smoke) ==================
@pytest.mark.parametrize(
    "model_name, expected_str",
    [
        ("openai/gpt-3.5-turbo", "LiteLLMWrapper(provider=openai, name=gpt-3.5-turbo)"),
        ("mock-model", "LiteLLMWrapper(name=mock-model)"),
    ],
)
def test_litellm_wrapper_str(model_name, expected_str, mock_litellm_wrapper):
    wrapper = mock_litellm_wrapper(model_name=model_name)
    assert str(wrapper) == expected_str


def test_litellm_wrapper_model_name_property(mock_litellm_wrapper):
    wrapper = mock_litellm_wrapper(model_name="mock-model")
    assert wrapper.model_name() == "mock-model"


# ================= END str/model_name (smoke) ==================


# ================= START completion methods tests =========================
class TestCompletionMethods:
    @pytest.mark.parametrize("method_name,is_async,stream", [
        ("_chat", False, False),
        ("_achat", True, False),
        ("_chat", False, True),
    ], ids=["sync_chat", "async_chat", "sync_chat_streaming"])
    @pytest.mark.asyncio
    async def test_chat(self, mock_litellm_wrapper, message_history, method_name, is_async, stream):
        content = "Mission: Impossible" if stream else "Mocked response"
        wrapper = mock_litellm_wrapper(content=content, stream=stream)
        method = getattr(wrapper, method_name)
        
        if is_async:
            result = await method(message_history)
        else:
            result = method(message_history)

        assert isinstance(result, (Response, Generator))
        
        
        if stream:
            result: Generator[str | Response, None, Response]
            for chunk in result:
                if isinstance(chunk, Response):
                    assert isinstance(chunk.message, AssistantMessage)
                    assert chunk.message.content == content
                elif not isinstance(chunk, str):
                    pytest.fail("Stream yielded non-string, non-Response chunk")
        else:
            result: Response
            assert isinstance(result.message, AssistantMessage)
            assert result.message.content == content

    @pytest.mark.parametrize("method_name,is_async,stream", [
        ("_structured", False, False),
        ("_astructured", True, False),
        ("_structured", False, True),
    ], ids=["sync_structured", "async_structured", "sync_structured_streaming"])
    @pytest.mark.asyncio
    async def test_structured(self, mock_litellm_wrapper, message_history, method_name, is_async, stream):
        class ExampleSchema(BaseModel):
            field: str

        wrapper = mock_litellm_wrapper(content='{"field": "VAL"}', stream=stream)
        method = getattr(wrapper, method_name)
        
        if is_async:
            result = await method(message_history, schema=ExampleSchema)
        else:
            result = method(message_history, schema=ExampleSchema)

        assert isinstance(result, (Response, Generator))
        
        
        if stream:
            for chunk in result:
                if isinstance(chunk, Response):
                    print(chunk)
                    assert isinstance(chunk.message, AssistantMessage)
                    assert isinstance(chunk.message.content, ExampleSchema)
                    assert chunk.message.content.field == "VAL"
                elif not isinstance(chunk, str):
                    pytest.fail("Stream yielded non-string, non-Response chunk")

            
        else:
            assert isinstance(result.message, AssistantMessage)
            assert isinstance(result.message.content, ExampleSchema)
            assert result.message.content.field == "VAL"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("method_name,is_async", [
        ("_structured", False),
        ("_astructured", True),
    ], ids=["sync_structured", "async_structured"])
    async def test_structured_schema_jsondecode_error(
        self, mock_litellm_wrapper, message_history, method_name, is_async
    ):
        class Schema(BaseModel):
            val: int

        with pytest.raises(JSONDecodeError):
            wrapper = mock_litellm_wrapper(content="Invalid JSON")
            method = getattr(wrapper, method_name)
            if is_async:
                result = await method(message_history, schema=Schema)
            else:
                result = method(message_history, schema=Schema)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("method_name,is_async", [
        ("_structured", False),
        ("_astructured", True),
    ], ids=["sync_structured", "async_structured"])
    async def test_structured_invalid_json_raises_llm_error(
        self, mock_litellm_wrapper, message_history, method_name, is_async
    ):
        class Schema(BaseModel):
            val: int
        with pytest.raises(LLMError, match="Structured LLM call failed"):
            wrapper = mock_litellm_wrapper(content='{"field": "VAL", "invalid": "json"}')
            method = getattr(wrapper, method_name)
            if is_async:
                result = await method(message_history, schema=Schema)
            else:
                result = method(message_history, schema=Schema)

    @pytest.mark.parametrize("method_name,is_async,stream", [
        ("_chat_with_tools", False, False),
        ("_achat_with_tools", True, False),
        ("_chat_with_tools", False, True),
        # ("_achat_with_tools", True, True),
    ], ids=[
        "sync_chat_with_tools",
        "async_chat_with_tools",
        "sync_chat_with_tools_streaming",
        # "async_chat_with_tools_streaming"
        ])
    @pytest.mark.asyncio
    async def test_chat_with_tools(
        self, mock_litellm_wrapper, message_history, tool, method_name, is_async, stream
    ):
        if stream:
            wrapper = mock_litellm_wrapper(
                content='[{"name": "tool_x", "arguments": {"foo": 1}, "identifier": "id123"}]',
                stream=True,
            )
        else:
            wrapper = mock_litellm_wrapper(
                content=None,
                tool_calls=[
                    litellm.ChatCompletionMessageToolCall(
                        function=litellm.Function(arguments='{"foo": 1}', name="tool_x"),
                        id="id123",
                        type="function",
                    )
                ],
            )
        
        method = getattr(wrapper, method_name)
        if is_async:
            result = await method(message_history, [tool])
        else:
            result = method(message_history, [tool])

        assert isinstance(result, (Response, Generator))

        
        if stream:  # no stream in case the llm requests tool
            for chunk in result:
                if isinstance(chunk, Response):
                    try:
                        calls = json.loads(chunk.message.content)
                        assert isinstance(calls, list)
                        assert calls[0]["name"] == "tool_x"
                        assert calls[0]["arguments"] == {"foo": 1}
                        assert calls[0]["identifier"] == "id123"
                    except Exception as e:
                        pytest.fail("Structured response did not match schema")
                elif not isinstance(chunk, str):
                    pytest.fail("Stream yielded non-string, non-Response chunk")

        else:
            assert isinstance(result.message, AssistantMessage)
            calls = result.message.content
            assert isinstance(calls, list)
            assert calls[0].name == "tool_x"
            assert calls[0].arguments == {"foo": 1}
            assert calls[0].identifier == "id123"

# ================= END completion methods tests =========================


from typing import Literal, TypeVar

from ...providers import ModelProvider
from .._model_exception_base import ModelNotFoundError
from ._provider_wrapper import ProviderLLMWrapper

_TStream = TypeVar("_TStream", Literal[True], Literal[False])


class HuggingFaceLLM(ProviderLLMWrapper[_TStream]):
    def _pre_init_provider_check(self, model_name):
        """called by __init__ before the super call in ProviderLLMWrapper"""
        # for huggingface models there is no good way of using `get_llm_provider` to check if the model is valid.
        # so we are just goinog to add `huggingface/` to the model name in case it is not there.
        # if the model name happens to be invalid, the error will be generated at runtime during `litellm.completion`. See `_litellm_wrapper.py`
        if model_name.startswith(self.model_provider().lower()):
            model_name = "/".join(model_name.split("/")[1:])
        try:
            assert len(model_name.split("/")) == 3, "Invalid model name"
        except AssertionError as e:
            raise ModelNotFoundError(
                reason=e.args[0],
                notes=[
                    "Model name must be of the format `huggingface/<provider>/<hf_org_or_user>/<hf_model>` or `<provider>/<hf_org_or_user>/<hf_model>`",
                    "We only support the huggingface Serverless Inference Provider Models.",
                    "Provider List: https://docs.litellm.ai/docs/providers",
                ],
            )
        return model_name

    def model_provider(self) -> ModelProvider:
        # TODO implement logic for all the possible providers attached the hugging face.
        return ModelProvider.HUGGINGFACE

    def _validate_tool_calling_support(self):
        # NOTE: special exception case for huggingface
        # Due to the wide range of huggingface models, `litellm.supports_function_calling` isn't always accurate.
        # so we are just going to skip the check and the error (if any) will be generated at runtime during `litellm.completion`.
        pass

    @classmethod
    def model_gateway(cls):
        return ModelProvider.HUGGINGFACE




import os
import asyncio
from typing import List

import pytest

import railtracks as rt
from railtracks.prebuilt import rag_node
from railtracks.rag.utils import read_file
from railtracks.llm import OpenAILLM
from railtracks.rag.rag_core import RAG, RAGConfig, SearchResult


pytestmark = pytest.mark.integration


def custom_rag_node(
    documents: List[str],
    embed_model: str = "text-embedding-3-small",
    token_count_model: str = "gpt-4o",
    chunk_size: int = 1000,
    chunk_overlap: int = 200,
):
    """Create a custom RAG node with specific configuration using the core RAG API."""
    rag_core = RAG(
        docs=documents,
        config=RAGConfig(
            embedding={"model": embed_model},
            store={},
            chunking={
                "chunk_size": chunk_size,
                "chunk_overlap": chunk_overlap,
                "model": token_count_model,
            },
        ),
    )
    rag_core.embed_all()

    def query(q: str, top_k: int = 1) -> SearchResult:
        return rag_core.search(q, top_k=top_k)

    return rt.function_node(query)


def test_simple_rag_example_e2e():
    retriever = rag_node(
        [
            "Steve likes apples and enjoys them as snacks",
            "John prefers bananas for their potassium content",
            "Alice loves oranges for vitamin C",
        ]
    )

    question = "What does Steve like?"
    results = asyncio.run(rt.call(retriever, question, top_k=3))

    context = "\n".join(
        f"Document {i+1} (score: {r.score:.4f}): {r.record.text}"
        for i, r in enumerate(results)
    )

    # Assertions: shape + relevance
    assert len(list(results)) == 3
    assert "Steve likes apples" in context
    # Contract checks for result entry and container
    first = list(results)[0]
    assert hasattr(first, "score") and isinstance(first.score, float)
    assert hasattr(first, "record") and hasattr(first.record, "text") and isinstance(first.record.text, str)
    assert hasattr(results, "to_list_of_texts")
    assert isinstance(results.to_list_of_texts(), list)


def test_rag_with_files_e2e(tmp_path, monkeypatch):
    # Prepare a docs directory with two files
    docs_dir = tmp_path / "docs"
    docs_dir.mkdir()
    (docs_dir / "faq.txt").write_text("FAQ: Common questions about working hours and tools.")
    (docs_dir / "policies.txt").write_text("Policies: Work from home allowed on Fridays.")

    # Use relative paths like users would
    monkeypatch.chdir(tmp_path)

    doc1_content = read_file("./docs/faq.txt")
    doc2_content = read_file("./docs/policies.txt")

    retriever = rag_node([doc1_content, doc2_content])

    question = "What is the work from home policy?"
    res = asyncio.run(rt.call(retriever, question, top_k=2))
    texts = res.to_list_of_texts()

    assert len(texts) == 2
    assert any("work from home" in t.lower() for t in texts)


def test_rag_topk_bounds_and_order_e2e():
    docs = [
        "Doc A: apples are red.",
        "Doc B: bananas are yellow.",
        "Doc C: oranges are orange.",
    ]
    retriever = rag_node(docs)

    # top_k larger than corpus should cap to the corpus size
    res_all = asyncio.run(rt.call(retriever, "What color are bananas?", top_k=10))
    assert len(list(res_all)) == len(docs)

    # top_k == 1 should return a single, relevant result
    res_one = asyncio.run(rt.call(retriever, "bananas", top_k=1))
    texts_one = res_one.to_list_of_texts()
    assert len(texts_one) == 1
    assert "bananas" in texts_one[0].lower()


@pytest.mark.skipif(not os.getenv("OPENAI_API_KEY"), reason="Requires OPENAI_API_KEY to run")
def test_rag_with_llm_e2e():
    retriever = rag_node(
        [
            "Our company policy requires all employees to work from home on Fridays",
            "Data security guidelines mandate encryption of all sensitive customer information",
            "Employee handbook states vacation requests need 2 weeks advance notice",
        ]
    )

    question = "What is the work from home policy?"
    search_result = asyncio.run(rt.call(retriever, question, top_k=2))
    context = "\n\n".join(search_result.to_list_of_texts())

    agent = rt.agent_node(llm=OpenAILLM("gpt-4o"))

    with rt.Session(context={"context": context, "question": question}):
        response = asyncio.run(
            rt.call(
                agent,
                user_input=(
                    "Based on the following context, please answer the question.\n"
                    "Context:\n"
                    f"{context}\n"
                    "Question:\n"
                    f"{question}\n"
                    "Answer based only on the context provided. "
                    'If the answer is not in the context, say "I don\'t know".'
                ),
            )
        )

    assert hasattr(response, "content")
    assert isinstance(response.content, str)
    assert response.content.strip() != ""
    # Heuristic: answer should reference "Friday" or "work from home"
    lc = response.content.lower()
    assert "friday" in lc or "work from home" in lc


@pytest.mark.skipif(not os.getenv("OPENAI_API_KEY"), reason="Requires OPENAI_API_KEY to run")
def test_custom_rag_node_build_and_query_e2e():
    retriever = custom_rag_node(
        [
            "Alpha team prefers morning meetings",
            "Beta team likes afternoon standsups",
            "Gamma team schedules evening retrospectives",
        ]
    )

    result = asyncio.run(rt.call(retriever, "When does Alpha team meet?", top_k=1))
    texts = result.to_list_of_texts()

    assert len(texts) == 1
    assert "alpha" in texts[0].lower()
    assert "morning" in texts[0].lower()


@pytest.mark.skipif(not os.getenv("OPENAI_API_KEY"), reason="Requires OPENAI_API_KEY to run")
def test_custom_rag_topk_and_contract_e2e():
    docs = [
        "Doc A: The onboarding checklist includes security training and device setup.",
        "Doc B: PTO requests require a 2-week prior notice.",
        "Doc C: Work from home is allowed every Friday, subject to manager approval.",
        "Doc D: Standups happen every morning at 9 AM.",
    ]
    retriever = custom_rag_node(docs, chunk_size=400, chunk_overlap=100)

    res = asyncio.run(rt.call(retriever, "When is WFH allowed?", top_k=2))
    items = list(res)
    assert 1 <= len(items) <= 2

    texts = res.to_list_of_texts()
    assert any("work from home" in t.lower() or "wfh" in t.lower() for t in texts)

    # Contract checks
    first = items[0]
    assert hasattr(first, "score") and isinstance(first.score, float)
    assert hasattr(first, "record") and hasattr(first.record, "text") and isinstance(first.record.text, str)


from typing import Type, Literal
from urllib import response

import pytest
import railtracks as rt
from railtracks.llm.providers import ModelProvider
from railtracks.llm.response import Response, MessageInfo
from railtracks.llm.content import Stream
from pydantic import BaseModel
from railtracks.llm.message import AssistantMessage, Message
import json


class MockLLM(rt.llm.ModelBase):
    def __init__(
        self,
        custom_response: str | None = None,
        requested_tool_calls: list[rt.llm.ToolCall] | None = None,
        stream: bool = False,
    ):
        """
        Creates a new instance of the MockLLM class.
        Args:
            custom_response_message (Message | None, optional): The custom response message to use for the LLM. Defaults to None.
        """
        super().__init__(stream=stream)
        self.custom_response = custom_response
        self.requested_tool_calls = requested_tool_calls
        self.mocked_message_info = MessageInfo(
            input_tokens=42,
            output_tokens=42,
            latency=1.42,
            model_name="MockLLM",
            total_cost=0.00042,
            system_fingerprint="fp_4242424242",
        )

    # ================================ HELPERS =================================================
    def _extract_pending_tool_results(self, messages):
        """
        Extract tool results from the end of the message history that need processing.
        """
        tool_results = []

        # Look backwards from the end for consecutive tool messages
        for message in reversed(messages):
            if message.role == "tool":
                tool_results.insert(0, message)  # Insert at beginning to maintain order
            else:
                break  # Stop at first non-tool message
        return tool_results

    # =======================================================================================

    # ================ Base responses (common for sync and async versions) ==================
    def _base_chat(self):
        if self.custom_response:
            assert isinstance(self.custom_response, str), "custom_response must be a string for terminal LLMs"
        return_message = self.custom_response or "mocked Message"
        # Streaming case
        if self.stream:
            def make_generator():
                    for char in return_message:
                        yield char

                    r = Response(
                        message=AssistantMessage(content=return_message),
                        message_info=self.mocked_message_info,
                    )
                    yield r 
                    return r
                
            return make_generator()
        
        # general case 
        return Response(
            message=AssistantMessage(return_message),
            message_info=self.mocked_message_info,
        )

    def _base_structured(self, messages, schema):
        class DummyStructured(BaseModel):
            dummy_attr: str = "mocked"

        if self.custom_response:
            response_model = schema(**json.loads(self.custom_response))
        else:
            response_model = DummyStructured()

        # Streaming case
        if self.stream:
            def make_generator():
                    for char in response_model.model_dump_json():
                        yield char

                    r = Response(
                        message=AssistantMessage(content=response_model),
                        message_info=self.mocked_message_info,
                    )
                    yield r 
                    return r
                
            return make_generator()

        return Response(
            message=AssistantMessage(response_model),
            message_info=self.mocked_message_info,
        )

    def _base_chat_with_tools(self, messages):
        tool_results = self._extract_pending_tool_results(messages)
        if tool_results:
            final_message = ""
            for tool_message in tool_results:
                tool_response = tool_message.content
                final_message += (
                    f"Tool {tool_response.name} returned: '{tool_response.result}'"
                    + "\n"
                )
            # Streaming case
            if self.stream:
                def make_generator():
                    for char in final_message:
                        yield char

                    r = Response(
                        message=AssistantMessage(content=final_message),
                        message_info=self.mocked_message_info,
                    )
                    yield r 
                    return r
                
                return make_generator()
            
            return Response(    # no changes in this response in case of streaming
                message=AssistantMessage(content=final_message),
                message_info=self.mocked_message_info,
            )
        else:
            return_message = self.requested_tool_calls or "mocked tool message"
            r = Response(    # no changes in this response in case of streaming
                message=AssistantMessage(return_message),
                message_info=self.mocked_message_info,
            )
            if self.stream:
                def tool_generator():
                    yield r
                    return r
                return tool_generator()
            else:
                return r


    # ==========================================================
    # Override all methods that make network calls with mocks
    async def _achat(self, messages, **kwargs):
        return self._base_chat()

    async def _astructured(self, messages, schema, **kwargs):
        return self._base_structured(messages, schema)

    async def _achat_with_tools(self, messages, tools, **kwargs):
        return self._base_chat_with_tools(messages, **kwargs)

    async def _astream_chat(self, messages, **kwargs):
        return self._base_chat()

    def _chat(self, messages, **kwargs):
        return self._base_chat()

    def _structured(self, messages, schema, **kwargs):
        return self._base_structured(messages, schema)

    def _chat_with_tools(self, messages, tools, **kwargs):
        return self._base_chat_with_tools(messages, **kwargs)

    def _stream_chat(self, messages, **kwargs):
        return self._base_chat()

    # ==========================================================

    # =====================================
    def model_name(self) -> str | None:
        return "MockLLM"

    @classmethod
    def model_gateway(cls) -> str | None:
        return "mock"
    
    def model_provider(self):
        return self.model_gateway()

    # =====================================


@pytest.fixture
def mock_llm() -> Type[MockLLM]:
    """
    Fixture to mock LLM methods with configurable responses.
    Pass a custom_response_message to override the message in all default responses.
    Usage:
        model = mock_model(
                    custom_response_message=r"custom")
                    requested_tool_calls=[ToolCall(name="secret_phrase", identifier="id_42424242", arguments={})]
                )
    """
    return MockLLM



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Optional, Sequence

import litellm

logger = logging.getLogger(__name__)


class BaseEmbeddingService(ABC):
    """
    Base class for embedding services.
    """

    @abstractmethod
    def embed(self, texts: Sequence[str], *, batch_size: int = 8) -> List[List[float]]:
        raise NotImplementedError("Subclasses must implement this method.")

    def __repr__(self):
        return f"{self.__class__.__name__}(model={self.model})"


class EmbeddingService(BaseEmbeddingService):
    """
    Embedding service that uses litellm to perform embedding tasks.
    """

    DEFAULT_MODEL = "text-embedding-3-small"

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        *,
        # --- litellm-specific kwargs ---
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: Optional[int] = 60,
        **litellm_extra: Any,
    ):
        """
        Initialize the embedding service.

        Args:
            model: Model name (OpenAI, TogetherAI, etc.)
            api_key: If None, taken from OPENAI_API_KEY env var.
            base_url: Override OpenAI base URL if using gateway / proxy.
            timeout: Per-request timeout in seconds.
            **litellm_extra: Any other args passed straight to litellm.embedding
                            (e.g. headers, organization, etc.)
        """
        self.model = model
        self.litellm_extra = {
            "api_key": api_key or None,
            "base_url": base_url,
            "timeout": timeout,
            **litellm_extra,
        }

    # ─────────────────────────────────────────────────────
    # PUBLIC API
    # ─────────────────────────────────────────────────────

    def embed(self, texts: Sequence[str], *, batch_size: int = 8) -> List[List[float]]:
        """
        Convenience wrapper to embed many short texts in one go.
        Accepts either a sequence of strings or a single string.
        """
        if isinstance(texts, (str, bytes)):
            texts = [str(texts)]

        if not isinstance(texts, Sequence):
            raise TypeError("texts must be a sequence of strings or a single string")

        if batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")

        vectors: List[List[float]] = []
        for i in range(0, len(texts), batch_size):
            batch = texts[i : i + batch_size]
            vectors.extend(self._embed_batch(batch))
        return vectors

    # ─────────────────────────────────────────────────────
    # INTERNAL HELPERS
    # ─────────────────────────────────────────────────────
    def _embed_batch(self, batch: Iterable[str]) -> List[List[float]]:
        """
        Low-level wrapper around litellm.embedding
        """
        batch_list = list(batch)
        if not batch_list:
            return []

        try:
            response = litellm.embedding(
                model=self.model,
                input=batch_list,
                **{k: v for k, v in self.litellm_extra.items() if v is not None},
            )
        except Exception as e:
            logger.exception("Embedding request failed: %s", e)
            raise

        # litellm typically returns a dict with 'data' or an object with .data
        data = getattr(response, "data", None)
        if data is None and isinstance(response, dict):
            data = response.get("data")

        if data is None:
            raise ValueError("Unexpected embedding response format: missing 'data'")

        # Normalize items to (index, embedding) and maintain order
        indexed: List[tuple[int, List[float]]] = []
        for i, item in enumerate(data):
            if isinstance(item, dict):
                idx = item.get("index", i)
                emb = item.get("embedding")
            else:
                idx = getattr(item, "index", i)
                emb = getattr(item, "embedding", None)

            if emb is None:
                raise ValueError(
                    "Unexpected embedding item format: missing 'embedding'"
                )

            indexed.append((idx, list(emb)))

        indexed.sort(key=lambda t: t[0])
        vectors = [emb for _, emb in indexed]
        return vectors



from abc import ABC, abstractmethod
from typing import Any, Dict

from pydantic import BaseModel


class HILMessage(BaseModel):
    content: str
    metadata: Dict[str, Any] | None = None


class HIL(ABC):
    @abstractmethod
    async def connect(self) -> None:
        """
        Creates or initializes the user interface component.
        """
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        """
        Disconnects the user interface component.
        """
        pass

    @abstractmethod
    async def send_message(
        self, content: HILMessage, timeout: float | None = None
    ) -> bool:
        """
        HIL uses this function to send a message to the user through the interface.

        Args:
            content: The message content to send.
            timeout: The maximum time in seconds to wait for the message to be sent.

        Returns:
            True if the message was sent successfully, False otherwise.
        """
        pass

    @abstractmethod
    async def receive_message(self, timeout: float | None = None) -> HILMessage | None:
        """
        HIL uses this function to wait for the user to provide input.

        This method should block until input is received or the timeout is reached.

        Args:
            timeout: The maximum time in seconds to wait for input.

        Returns:
            The user input if received within the timeout period, None otherwise.
        """
        pass



from __future__ import annotations

import os
from typing import Callable, Coroutine

from railtracks.utils.logging.config import AllowableLogLevels, str_to_log_level


class ExecutorConfig:
    def __init__(
        self,
        *,
        timeout: float = 150.0,
        end_on_error: bool = False,
        logging_setting: AllowableLogLevels = "INFO",
        log_file: str | os.PathLike | None = None,
        broadcast_callback: (
            Callable[[str], None] | Callable[[str], Coroutine[None, None, None]] | None
        ) = None,
        prompt_injection: bool = True,
        save_state: bool = True,
    ):
        """
        ExecutorConfig is special configuration object designed to allow customization of the executor in the RT system.

        Args:
            timeout (float): The maximum number of seconds to wait for a response to your top level request
            end_on_error (bool): If true, the executor will stop execution when an exception is encountered.
            logging_setting (AllowableLogLevels): The setting for the level of logging you would like to have.
            log_file (str | os.PathLike | None): The file to which the logs will be written. If None, no file will be created.
            broadcast_callback (Callable or Coroutine): A function or coroutine that will handle streaming messages.
            prompt_injection (bool): If true, prompts can be injected with global context
            save_state (bool): If true, the state of the executor will be saved to disk.
        """
        self.timeout = timeout
        self.end_on_error = end_on_error
        self.logging_setting = logging_setting
        self.subscriber = broadcast_callback
        self.log_file = log_file
        self.prompt_injection = prompt_injection
        self.save_state = save_state

    @property
    def logging_setting(self) -> AllowableLogLevels:
        return self._logging_setting

    @logging_setting.setter
    def logging_setting(self, value: AllowableLogLevels):
        if value not in str_to_log_level:
            raise ValueError(
                f"logging_setting must be one of {str_to_log_level}, got {value}"
            )
        self._logging_setting: AllowableLogLevels = value

    def precedence_overwritten(
        self,
        *,
        timeout: float | None = None,
        end_on_error: bool | None = None,
        logging_setting: AllowableLogLevels | None = None,
        log_file: str | os.PathLike | None = None,
        subscriber: (
            Callable[[str], None] | Callable[[str], Coroutine[None, None, None]] | None
        ) = None,
        prompt_injection: bool | None = None,
        save_state: bool | None = None,
    ):
        """
        If any of the parameters are provided (not None), it will create a new update the current instance with the new values and return a deep copied reference to it.
        """
        return ExecutorConfig(
            timeout=timeout if timeout is not None else self.timeout,
            end_on_error=end_on_error
            if end_on_error is not None
            else self.end_on_error,
            logging_setting=logging_setting
            if logging_setting is not None
            else self.logging_setting,
            log_file=log_file if log_file is not None else self.log_file,
            broadcast_callback=subscriber
            if subscriber is not None
            else self.subscriber,
            prompt_injection=prompt_injection
            if prompt_injection is not None
            else self.prompt_injection,
            save_state=save_state if save_state is not None else self.save_state,
        )

    def __repr__(self):
        return (
            f"ExecutorConfig(timeout={self.timeout}, end_on_error={self.end_on_error}, "
            f"logging_setting={self.logging_setting}, log_file={self.log_file}, "
            f"prompt_injection={self.prompt_injection}, "
            f"save_state={self.save_state})"
        )



from __future__ import annotations

from typing import List

from .message import Message, Role


class MessageHistory(List[Message]):
    """
    A basic object that represents a history of messages. The object has all the same capability as a list such as
    `.remove()`, `.append()`, etc.
    """

    def __str__(self):
        return "\n".join([str(message) for message in self])

    def removed_system_messages(self) -> MessageHistory:
        """
        Returns a new MessageHistory object with all SystemMessages removed.
        """
        return MessageHistory([msg for msg in self if msg.role != Role.system])



import pytest
from railtracks.llm.tools.parameters.union_parameter import UnionParameter
from railtracks.llm.tools.parameters._base import Parameter, ParameterType

def make_dummy_param(name, param_type):
    class Dummy(Parameter):
        def to_json_schema(self):
            return {"type": param_type}
    return Dummy(name, param_type=param_type)

def test_union_parameter_basic():
    p1 = make_dummy_param("foo", "string")
    p2 = make_dummy_param("bar", "integer")
    union = UnionParameter("baz", options=[p1, p2], description="desc", required=True, default_present=True, default=123)
    schema = union.to_json_schema()
    assert "anyOf" in schema
    assert {"type": "string"} in schema["anyOf"]
    assert {"type": "integer"} in schema["anyOf"]
    assert schema["description"] == "desc"
    assert schema["default"] == 123

def test_union_parameter_no_nested_union():
    p1 = make_dummy_param("foo", "string")
    with pytest.raises(TypeError):
        UnionParameter("baz", options=[p1, UnionParameter("bad", options=[p1])])

def test_union_parameter_repr():
    p1 = make_dummy_param("foo", "string")
    p2 = make_dummy_param("bar", "integer")
    union = UnionParameter("baz", options=[p1, p2])
    s = repr(union)
    assert "UnionParameter" in s
    assert "baz" in s
